# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import copy
from typing import ClassVar, List, Optional, Tuple

import attr

from geti_sdk.data_models.enums.domain import Domain


@attr.define
class LabelSource:
    """
    Representation of a source for a ScoredLabel in GETi

    :var user_id: ID of the user who assigned the label, if any
    :var model_id: ID of the model which generated the label, if any
    :var model_storage_id: ID of the model storage to which the model belongs
    """

    user_id: Optional[str] = None
    model_id: Optional[str] = None
    model_storage_id: Optional[str] = None


@attr.define
class Label:
    """
    Representation of a Label in GETi.

    :var name: Name of the label
    :var id: Unique database ID of the label
    :var color: Color (in hex representation) of the label
    :var group: Name of the label group to which the label belongs
    :var is_empty: True if the label represents an empty label, False otherwise
    :var parent_id: Optional name of the parent label, if any
    """

    _identifier_fields: ClassVar[List[str]] = ["id", "hotkey"]
    _GET_only_fields: ClassVar[List[str]] = ["is_empty", "is_anomalous"]

    name: str
    color: str
    group: str
    is_empty: bool
    hotkey: str = ""
    domain: Optional[Domain] = None
    id: Optional[str] = None
    parent_id: Optional[str] = None
    is_anomalous: Optional[bool] = None

    def __key(self) -> Tuple[str, str]:
        """
        Return a tuple representing the key of the label.

        The key is a tuple containing the name and color of the label.

        :return: A tuple representing the key of the label.
        """
        return (self.name, self.color)

    def __hash__(self) -> int:
        """
        Calculate the hash value of the object.

        :return: The hash value of the object.
        """
        return hash(self.__key())

    def prepare_for_post(self) -> None:
        """
        Set all fields to None that are not valid for making a POST request to the
        /projects endpoint.

        :return:
        """
        for field_name in self._GET_only_fields:
            setattr(self, field_name, None)

    @property
    def color_tuple(self) -> Tuple[int, int, int]:
        """
        Return the color of the label as an RGB tuple.

        :return:
        """
        hex_color_str = copy.deepcopy(self.color).strip("#")
        return tuple(int(hex_color_str[i : i + 2], 16) for i in (0, 2, 4))


@attr.define
class ScoredLabel:
    """
    Representation of a Label with an assigned probability in GETi.

    :var name: Name of the label
    :var id: Unique database ID of the label
    :var color: Color (in hex representation) of the label
    :var probability:
    :var source:
    """

    _identifier_fields: ClassVar[List[str]] = ["id"]

    probability: float
    name: Optional[str] = None
    color: Optional[str] = None
    id: Optional[str] = None
    source: Optional[LabelSource] = None

    @property
    def color_tuple(self) -> Tuple[int, int, int]:
        """
        Return the color of the label as an RGB tuple.

        :return:
        """
        hex_color_str = copy.deepcopy(self.color).strip("#")
        return tuple(int(hex_color_str[i : i + 2], 16) for i in (0, 2, 4))

    @classmethod
    def from_label(cls, label: Label, probability: float) -> "ScoredLabel":
        """
        Create a ScoredLabel instance from an input Label and probability score.

        :param label: Label to convert to ScoredLabel
        :param probability: probability score for the label
        :return: ScoredLabel instance corresponding to `label` and `probability`
        """
        return ScoredLabel(
            name=label.name, probability=probability, color=label.color, id=label.id
        )

    def __key(self) -> Tuple[str, str]:
        """
        Return a tuple representing the key of the ScoredLabel.

        The key is a tuple containing the name and color of the scored label.

        :return: A tuple representing the key of the label.
        """
        return (self.name, self.color)

    def __hash__(self) -> int:
        """
        Calculate the hash value of the object.

        :return: The hash value of the object.
        """
        return hash(self.__key())



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from enum import Enum


class JobType(Enum):
    """
    Enum representing the type of a job on the Intel® Geti™ cluster.
    """

    UNDEFINED = "undefined"
    TRAIN = "train"
    INFERENCE = "inference"
    RECONSTRUCT_VIDEO = "reconstruct_video"
    EVALUATE = "evaluate"
    OPTIMIZATION = "optimization"
    OPTIMIZE_POT = "optimize_pot"
    OPTIMIZE_NNCF = "optimize_nncf"
    TEST = "test"
    PREPARE_IMPORT_TO_NEW_PROJECT = "prepare_import_to_new_project"
    PERFORM_IMPORT_TO_NEW_PROJECT = "perform_import_to_new_project"

    def __str__(self) -> str:
        """
        Return the string representation of the JobType instance.
        """
        return self.value



# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
"""
PostInferenceHooks are used to construct pipelines for e.g. data collection or other
actions based on inference outcomes.
"""
from .actions import FileSystemDataCollection, GetiDataCollection, HttpRequestAction
from .post_inference_hook import PostInferenceHook
from .triggers import (
    AlwaysTrigger,
    ConfidenceTrigger,
    EmptyLabelTrigger,
    LabelTrigger,
    ObjectCountTrigger,
)

__all__ = [
    "PostInferenceHook",
    "ConfidenceTrigger",
    "LabelTrigger",
    "EmptyLabelTrigger",
    "ObjectCountTrigger",
    "AlwaysTrigger",
    "GetiDataCollection",
    "FileSystemDataCollection",
    "HttpRequestAction",
]



# INTEL CONFIDENTIAL
#
# Copyright (C) 2024 Intel Corporation
#
# This software and the related documents are Intel copyrighted materials, and
# your use of them is governed by the express license under which they were provided to
# you ("License"). Unless the License provides otherwise, you may not use, modify, copy,
# publish, distribute, disclose or transmit this software or the related documents
# without Intel's prior written permission.
#
# This software and the related documents are provided as is,
# with no express or implied warranties, other than those that are expressly stated
# in the License.

import json
import logging
from typing import Any, Dict, List

import numpy as np
from model_api.models.utils import Detection

from geti_sdk.data_models.model import Model

# from sc_sdk.entities.model import Model

logger = logging.getLogger(__name__)


def detection2array(detections: List[Detection]) -> np.ndarray:
    """
    Convert list of OpenVINO Detection to a numpy array.

    :param detections: list of OpenVINO Detection containing [score, id, xmin, ymin, xmax, ymax]
    :return: numpy array with [label, confidence, x1, y1, x2, y2]
    """
    scores = np.empty((0, 1), dtype=np.float32)
    labels = np.empty((0, 1), dtype=np.uint32)
    boxes = np.empty((0, 4), dtype=np.float32)
    for det in detections:
        if (det.xmax - det.xmin) * (det.ymax - det.ymin) < 1.0:
            continue
        scores = np.append(scores, [[det.score]], axis=0)
        labels = np.append(labels, [[det.id]], axis=0)
        boxes = np.append(
            boxes,
            [[float(det.xmin), float(det.ymin), float(det.xmax), float(det.ymax)]],
            axis=0,
        )
    return np.concatenate((labels, scores, boxes), -1)


def get_detection_inferencer_configuration(model: Model) -> dict:
    """
    Get detection configuration from the model.

    :param model: (Geti) Model to get the detection configuration from
    :return: dict representing the detection configuration
    """
    config = json.loads(model.get_data("config.json"))
    _flatten_config_values(config)

    configuration = {}
    if config["postprocessing"].get("result_based_confidence_threshold", False):
        configuration["confidence_threshold"] = float(
            np.frombuffer(model.get_data("confidence_threshold"), dtype=np.float32)[0]
        )
    configuration["use_ellipse_shapes"] = config["postprocessing"].get(
        "use_ellipse_shapes", False
    )

    logger.info(f"Detection inferencer configuration: {configuration}")
    return configuration


def _flatten_config_values(config: Dict[str, Any]) -> None:
    """
    Extract the "value" field from any nested config.

    Flattening the structure of the config dictionary. The original config dictionary is modified in-place.

    :param config: config dictionary
    """
    for key, value in config.items():
        if isinstance(value, dict):
            if "value" in value:
                config[key] = value["value"]
            else:
                _flatten_config_values(value)



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from enum import Enum


class TaskType(Enum):
    """
    Enum representing the different task types in Intel® Geti™ projects.
    """

    DETECTION = "detection"
    SEGMENTATION = "segmentation"
    CLASSIFICATION = "classification"
    ANOMALY_CLASSIFICATION = "anomaly_classification"
    ANOMALY_DETECTION = "anomaly_detection"
    ANOMALY_SEGMENTATION = "anomaly_segmentation"
    INSTANCE_SEGMENTATION = "instance_segmentation"
    ROTATED_DETECTION = "rotated_detection"
    DATASET = "dataset"
    CROP = "crop"

    def __str__(self) -> str:
        """
        Return the string representation of the TaskType instance.

        :return: string containing the task type
        """
        return self.value

    @property
    def is_trainable(self) -> bool:
        """
        Return True if a task of this TaskType is trainable, False otherwise.

        :return:
        """
        return self not in NON_TRAINABLE_TASK_TYPES

    @property
    def is_global(self) -> bool:
        """
        Return True if a task of this TaskType produces global labels, False otherwise.

        :return:
        """
        return self in GLOBAL_TASK_TYPES

    @property
    def is_local(self) -> bool:
        """
        Return True if a task of this TaskType produces local labels, False otherwise.

        :return:
        """
        return self not in GLOBAL_TASK_TYPES and self not in NON_TRAINABLE_TASK_TYPES

    @property
    def is_anomaly(self) -> bool:
        """
        Return True if a task of this TaskType is an anomaly task, False otherwise.

        :return:
        """
        return self in ANOMALY_TASK_TYPES

    @property
    def is_segmentation(self) -> bool:
        """
        Return True if a task of this TaskType is a segmentation task, False otherwise.

        :return:
        """
        return self in SEGMENTATION_TASK_TYPES

    @property
    def is_detection(self) -> bool:
        """
        Return True if a task of this TaskType is a detection task, False otherwise.

        :return:
        """
        return self in DETECTION_TASK_TYPES

    @classmethod
    def from_domain(cls, domain):
        """
        Instantiate a :py:class:`~geti_sdk.data_models.enums.task_type.TaskType`
        from a given :py:class:`~geti_sdk.data_models.enums.domain.Domain`.

        :param domain: domain to get the TaskType for
        :return: TaskType instance corresponding to the `domain`
        """
        return cls[domain.name]


NON_TRAINABLE_TASK_TYPES = [TaskType.DATASET, TaskType.CROP]

ANOMALY_TASK_TYPES = [
    TaskType.ANOMALY_CLASSIFICATION,
    TaskType.ANOMALY_DETECTION,
    TaskType.ANOMALY_SEGMENTATION,
]

GLOBAL_TASK_TYPES = [TaskType.CLASSIFICATION, TaskType.ANOMALY_CLASSIFICATION]

SEGMENTATION_TASK_TYPES = [
    TaskType.SEGMENTATION,
    TaskType.ANOMALY_SEGMENTATION,
    TaskType.INSTANCE_SEGMENTATION,
]

DETECTION_TASK_TYPES = [
    TaskType.DETECTION,
    TaskType.ROTATED_DETECTION,
    TaskType.ANOMALY_DETECTION,
]



# INTEL CONFIDENTIAL
#
# Copyright (C) 2024 Intel Corporation
#
# This software and the related documents are Intel copyrighted materials, and
# your use of them is governed by the express license under which they were provided to
# you ("License"). Unless the License provides otherwise, you may not use, modify, copy,
# publish, distribute, disclose or transmit this software or the related documents
# without Intel's prior written permission.
#
# This software and the related documents are provided as is,
# with no express or implied warranties, other than those that are expressly stated
# in the License.

"""Module implements the InferenceResultsToPredictionConverter class."""

import abc
from typing import Any, Dict, List, NamedTuple, Optional, Tuple, Union

import cv2
import numpy as np
from model_api.models.utils import (
    AnomalyResult,
    ClassificationResult,
    DetectionResult,
    ImageResultWithSoftPrediction,
    InstanceSegmentationResult,
)

from geti_sdk.data_models.annotations import Annotation
from geti_sdk.data_models.enums.domain import Domain
from geti_sdk.data_models.label import ScoredLabel
from geti_sdk.data_models.label_schema import LabelSchema
from geti_sdk.data_models.predictions import Prediction
from geti_sdk.data_models.shapes import (
    Ellipse,
    Point,
    Polygon,
    Rectangle,
    RotatedRectangle,
)
from geti_sdk.deployment.predictions_postprocessing.utils.detection_utils import (
    detection2array,
)
from geti_sdk.deployment.predictions_postprocessing.utils.segmentation_utils import (
    create_annotation_from_segmentation_map,
)


class InferenceResultsToPredictionConverter(metaclass=abc.ABCMeta):
    """Interface for the converter"""

    @property
    @abc.abstractmethod
    def domain(self) -> Domain:
        """
        Return the domain for which the converter applies

        :return: The task domain for which the label converter applies
        """
        raise NotImplementedError

    @abc.abstractmethod
    def convert_to_prediction(self, predictions: NamedTuple, **kwargs) -> Prediction:
        """
        Convert raw predictions to Prediction format.

        :param predictions: raw predictions from inference
        :return: Prediction object containing the shapes obtained from the raw predictions.
        """
        raise NotImplementedError


class ClassificationToPredictionConverter(InferenceResultsToPredictionConverter):
    """
    Converts ModelAPI Classification predictions to Prediction object.

    :param label_schema: LabelSchema containing the label info of the task
    """

    domain = Domain.CLASSIFICATION

    def __init__(self, label_schema: LabelSchema):
        all_labels = label_schema.get_labels(include_empty=True)
        # add empty labels if only one non-empty label exits
        non_empty_labels = [label for label in all_labels if not label.is_empty]
        self.labels = all_labels if len(non_empty_labels) == 1 else non_empty_labels
        # get the first empty label
        self.empty_label = next((label for label in all_labels if label.is_empty), None)
        multilabel = len(label_schema.get_groups(False)) > 1
        multilabel = multilabel and len(label_schema.get_groups(False)) == len(
            label_schema.get_labels(include_empty=False)
        )
        self.hierarchical = not multilabel and len(label_schema.get_groups(False)) > 1

        self.label_schema = label_schema

    def convert_to_prediction(
        self,
        predictions: ClassificationResult,
        image_shape: Tuple[int, int, int],
        **kwargs,
    ) -> Prediction:  # noqa: ARG003
        """
        Convert ModelAPI ClassificationResult predictions to Prediction object.

        :param predictions: classification labels represented in ModelAPI format (label_index, label_name, confidence)
        :return: Prediction object with corresponding label
        """
        labels = []
        for label in predictions.top_labels:
            labels.append(
                ScoredLabel.from_label(self.labels[label[0]], float(label[-1]))
            )

        if not labels and self.empty_label:
            labels = [ScoredLabel.from_label(self.empty_label, probability=1.0)]

        annotations = Annotation(
            shape=Rectangle.generate_full_box(image_shape[1], image_shape[0]),
            labels=labels,
        )
        return Prediction([annotations])


class DetectionToPredictionConverter(InferenceResultsToPredictionConverter):
    """
    Converts ModelAPI Detection objects to Prediction object.

    :param label_schema: LabelSchema containing the label info of the task
    :param configuration: optional model configuration setting
    """

    domain = Domain.DETECTION

    def __init__(
        self, label_schema: LabelSchema, configuration: Optional[Dict[str, Any]] = None
    ):
        self.labels = label_schema.get_labels(include_empty=False)
        self.label_map = dict(enumerate(self.labels))
        self.use_ellipse_shapes = False
        self.confidence_threshold = 0.0
        if configuration is not None:
            if "use_ellipse_shapes" in configuration:
                self.use_ellipse_shapes = configuration["use_ellipse_shapes"]
            if "confidence_threshold" in configuration:
                self.confidence_threshold = configuration["confidence_threshold"]

    def convert_to_prediction(
        self, predictions: DetectionResult, **kwargs
    ) -> Prediction:
        """
        Convert ModelAPI DetectionResult predictions to Prediction object.

        :param predictions: detection represented in ModelAPI format (label, confidence, x1, y1, x2, y2).

        _Note:
            - `label` can be any integer that can be mapped to `self.labels`
            - `confidence` should be a value between 0 and 1
            - `x1`, `x2`, `y1` and `y2` are expected to be in pixel
        :return: Prediction object containing the boxes obtained from the prediction
        """
        detections = detection2array(predictions.objects)

        annotations = []
        if (
            len(detections)
            and detections.shape[1:] < (6,)
            or detections.shape[1:] > (7,)
        ):
            raise ValueError(
                f"Shape of prediction is not expected, expected (n, 7) or (n, 6) but got {detections.shape}"
            )

        for detection in detections:
            # Some OpenVINO models use an output shape of [7,]
            # If this is the case, skip the first value as it is not used
            _detection = detection[1:] if detection.shape == (7,) else detection

            label = int(_detection[0])
            confidence = _detection[1]
            scored_label = ScoredLabel.from_label(self.label_map[label], confidence)
            coords = _detection[2:]
            shape: Union[Ellipse, Rectangle]

            if confidence < self.confidence_threshold:
                continue

            bbox_width = coords[2] - coords[0]
            bbox_height = coords[3] - coords[1]
            if self.use_ellipse_shapes:
                shape = Ellipse(coords[0], coords[1], bbox_width, bbox_height)
            else:
                shape = Rectangle(coords[0], coords[1], bbox_width, bbox_height)

            annotation = Annotation(shape=shape, labels=[scored_label])
            annotations.append(annotation)
        return Prediction(annotations)


class RotatedRectToPredictionConverter(DetectionToPredictionConverter):
    """
    Converts ModelAPI Rotated Detection objects to Prediction.

    :param label_schema: LabelSchema containing the label info of the task
    """

    domain = Domain.ROTATED_DETECTION

    def convert_to_prediction(
        self, predictions: InstanceSegmentationResult, **kwargs
    ) -> Prediction:
        """
        Convert ModelAPI instance segmentation predictions to a rotated bounding box annotation format.

        :param predictions: segmentation represented in ModelAPI format
        :return: Prediction object containing the rotated boxes obtained from the segmentation contours
        :raises ValueError: if metadata is missing from the preprocess step
        """
        annotations = []
        shape: Union[RotatedRectangle, Ellipse]
        for obj in predictions.segmentedObjects:
            if obj.score < self.confidence_threshold:
                continue
            if self.use_ellipse_shapes:
                shape = Ellipse(
                    obj.xmin, obj.ymin, obj.xmax - obj.xmin, obj.ymax - obj.ymin
                )
                annotations.append(
                    Annotation(
                        shape=shape,
                        labels=[
                            ScoredLabel.from_label(
                                self.labels[int(obj.id) - 1], float(obj.score)
                            )
                        ],
                    )
                )
            else:
                mask = obj.mask.astype(np.uint8)
                contours, hierarchies = cv2.findContours(
                    mask, cv2.RETR_CCOMP, cv2.CHAIN_APPROX_SIMPLE
                )
                if hierarchies is None:
                    continue
                for contour, hierarchy in zip(contours, hierarchies[0]):
                    if hierarchy[3] != -1:
                        continue
                    if len(contour) <= 2 or cv2.contourArea(contour) < 1.0:
                        continue
                    points = [
                        Point(
                            x=point[0],
                            y=point[1],
                        )
                        for point in cv2.boxPoints(cv2.minAreaRect(contour))
                    ]
                    shape = Polygon(points=points)
                    annotations.append(
                        Annotation(
                            shape=RotatedRectangle.from_polygon(shape),
                            labels=[
                                ScoredLabel.from_label(
                                    self.labels[int(obj.id) - 1], float(obj.score)
                                )
                            ],
                        )
                    )
        return Prediction(annotations)


class MaskToAnnotationConverter(InferenceResultsToPredictionConverter):
    """Converts DetectionBox Predictions ModelAPI to Prediction object."""

    domain = Domain.INSTANCE_SEGMENTATION

    def __init__(
        self, label_schema: LabelSchema, configuration: Optional[Dict[str, Any]] = None
    ):
        self.labels = label_schema.get_labels(include_empty=False)
        self.use_ellipse_shapes = False
        self.confidence_threshold = 0.0
        if configuration is not None:
            if "use_ellipse_shapes" in configuration:
                self.use_ellipse_shapes = configuration["use_ellipse_shapes"]
            if "confidence_threshold" in configuration:
                self.confidence_threshold = configuration["confidence_threshold"]

    def convert_to_prediction(
        self, predictions: Any, **kwargs: Dict[str, Any]
    ) -> Prediction:
        """
        Convert predictions to Prediction object.

        :param predictions: Raw predictions from the model.
        :return: Prediction object.
        """
        annotations = []
        shape: Union[Polygon, Ellipse]
        for obj in predictions.segmentedObjects:
            if obj.score < self.confidence_threshold:
                continue
            if self.use_ellipse_shapes:
                shape = shape = Ellipse(
                    obj.xmin, obj.ymin, obj.xmax - obj.xmin, obj.ymax - obj.ymin
                )
                annotations.append(
                    Annotation(
                        shape=shape,
                        labels=[
                            ScoredLabel.from_label(
                                self.labels[int(obj.id) - 1], float(obj.score)
                            )
                        ],
                    )
                )
            else:
                mask = obj.mask.astype(np.uint8)
                contours, hierarchies = cv2.findContours(
                    mask, cv2.RETR_CCOMP, cv2.CHAIN_APPROX_SIMPLE
                )
                if hierarchies is None:
                    continue
                for contour, hierarchy in zip(contours, hierarchies[0]):
                    if hierarchy[3] != -1:
                        continue
                    if len(contour) <= 2 or cv2.contourArea(contour) < 1.0:
                        continue
                    contour = list(contour)
                    points = [
                        Point(
                            x=point[0][0],
                            y=point[0][1],
                        )
                        for point in contour
                    ]
                    shape = Polygon(points=points)
                    annotations.append(
                        Annotation(
                            shape=shape,
                            labels=[
                                ScoredLabel.from_label(
                                    self.labels[int(obj.id) - 1], float(obj.score)
                                )
                            ],
                        )
                    )
        return Prediction(annotations)


class SegmentationToPredictionConverter(InferenceResultsToPredictionConverter):
    """
    Converts ModelAPI Segmentation objects to Prediction object.

    :param label_schema: LabelSchema containing the label info of the task
    """

    domain = Domain.SEGMENTATION

    def __init__(self, label_schema: LabelSchema):
        self.labels = label_schema.get_labels(include_empty=False)
        # NB: index=0 is reserved for the background label
        self.label_map = dict(enumerate(self.labels, 1))

    def convert_to_prediction(
        self, predictions: ImageResultWithSoftPrediction, **kwargs  # noqa: ARG002
    ) -> Prediction:
        """
        Convert ModelAPI instance segmentation predictions to Prediction object.

        :param predictions: segmentation represented in ModelAPI format
        :return: Prediction object containing the contour polygon obtained from the segmentation
        """
        annotations = create_annotation_from_segmentation_map(
            hard_prediction=predictions.resultImage,
            soft_prediction=predictions.soft_prediction,
            label_map=self.label_map,
        )
        return Prediction(annotations)


class AnomalyToPredictionConverter(InferenceResultsToPredictionConverter):
    """
    Convert ModelAPI AnomalyResult predictions to Prediction object.

    :param label_schema: LabelSchema containing the label info of the task
    """

    domain = Domain.ANOMALY

    def __init__(self, label_schema: LabelSchema):
        self.labels = label_schema.get_labels(include_empty=False)
        self.normal_label = next(
            label for label in self.labels if not label.is_anomalous
        )
        self.anomalous_label = next(
            label for label in self.labels if label.is_anomalous
        )
        self.domain = self.anomalous_label.domain

    def convert_to_prediction(
        self, predictions: AnomalyResult, image_shape: Tuple[int], **kwargs
    ) -> Prediction:  # noqa: ARG002
        """
        Convert ModelAPI AnomalyResult predictions to sc_sdk annotations.

        :param predictions: anomaly result represented in ModelAPI format (same for all anomaly tasks)
        :return: Prediction object based on the specific anomaly task:
            - Classification: single label (normal or anomalous).
            - Segmentation: contour polygon representing the segmentation.
            - Detection: predicted bounding boxes.
        """
        pred_label = predictions.pred_label
        label = (
            self.anomalous_label
            if pred_label in ("Anomaly", "Anomalous")
            else self.normal_label
        )
        annotations: List[Annotation] = []
        if self.domain == Domain.ANOMALY_CLASSIFICATION:
            scored_label = ScoredLabel.from_label(
                label=label, probability=float(predictions.pred_score)
            )
            annotations = [
                Annotation(
                    shape=Rectangle.generate_full_box(*image_shape[1::-1]),
                    labels=[scored_label],
                )
            ]
        elif self.domain == Domain.ANOMALY_SEGMENTATION:
            annotations = create_annotation_from_segmentation_map(
                hard_prediction=predictions.pred_mask,
                soft_prediction=predictions.anomaly_map.squeeze(),
                label_map={0: self.normal_label, 1: self.anomalous_label},
            )
        elif self.domain == Domain.ANOMALY_DETECTION:
            for box in predictions.pred_boxes:
                annotations.append(
                    Annotation(
                        shape=Rectangle(
                            box[0], box[1], box[2] - box[0], box[3] - box[1]
                        ),
                        labels=[
                            ScoredLabel.from_label(
                                label=self.anomalous_label,
                                probability=predictions.pred_score,
                            )
                        ],
                    )
                )
        else:
            raise ValueError(
                f"Cannot convert predictions for task '{self.domain.name}'. Only Anomaly tasks are supported."
            )
        if not annotations:
            scored_label = ScoredLabel.from_label(
                label=self.normal_label, probability=1.0
            )
            annotations = [
                Annotation(
                    labels=[scored_label],
                    shape=Rectangle.generate_full_box(*image_shape[1::-1]),
                )
            ]
        return Prediction(annotations)


class ConverterFactory:
    """
    Factory class for creating inference result to prediction converters based on the model's task.
    """

    @staticmethod
    def create_converter(
        label_schema: LabelSchema,
        configuration: Optional[Dict[str, Any]] = None,
    ) -> InferenceResultsToPredictionConverter:
        """
        Create the appropriate inferencer object according to the model's task.

        :param label_schema: The label schema containing the label info of the task.
        :param configuration: Optional configuration for the converter. Defaults to None.
        :return: The created inference result to prediction converter.
        :raises ValueError: If the task type cannot be determined from the label schema.
        """
        domain = ConverterFactory._get_labels_domain(label_schema)

        if domain == Domain.CLASSIFICATION:
            return ClassificationToPredictionConverter(label_schema)
        if domain == Domain.DETECTION:
            return DetectionToPredictionConverter(label_schema, configuration)
        if domain == Domain.SEGMENTATION:
            return SegmentationToPredictionConverter(label_schema)
        if domain == Domain.ROTATED_DETECTION:
            return RotatedRectToPredictionConverter(label_schema, configuration)
        if domain == Domain.INSTANCE_SEGMENTATION:
            return MaskToAnnotationConverter(label_schema, configuration)
        if domain in (
            Domain.ANOMALY_CLASSIFICATION,
            Domain.ANOMALY_SEGMENTATION,
            Domain.ANOMALY_DETECTION,
        ):
            return AnomalyToPredictionConverter(label_schema)
        raise ValueError(f"Cannot create inferencer for task type '{domain.name}'.")

    @staticmethod
    def _get_labels_domain(label_schema: LabelSchema) -> Domain:
        """
        Return the domain (task type) associated with the model's labels.

        :param label_schema: The label schema containing the label info of the task.
        :return: The domain of the task.
        :raises ValueError: If the task type cannot be determined from the label schema.
        """
        try:
            return next(
                label.domain for label in label_schema.get_labels(include_empty=False)
            )
        except StopIteration:
            raise ValueError("Cannot determine the task for the model")



# Copyright (C) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
from typing import List, Optional

import attr

from geti_sdk.data_models.enums import OptimizationType
from geti_sdk.data_models.utils import (
    str_to_datetime,
    str_to_enum_converter,
    str_to_task_type,
)


@attr.define()
class DatasetInfo:
    """
    Container for dataset information, specific to test datasets
    """

    id: str
    is_deleted: bool
    n_frames: int
    n_images: int
    n_samples: int
    name: str


@attr.define()
class JobInfo:
    """
    Container for job information, specific to model testing jobs
    """

    id: str
    status: str

    @property
    def is_done(self) -> bool:
        """
        Return True if the testing job has finished, False otherwise

        :return: True for a finished job, False otherwise
        """
        return self.status.lower() == "done"


@attr.define()
class ModelInfo:
    """
    Container for information related to the model, specific for model tests
    """

    group_id: str
    id: str
    n_labels: int
    task_type: str = attr.field(converter=str_to_task_type)
    template_id: str
    optimization_type: str = attr.field(
        converter=str_to_enum_converter(OptimizationType)
    )
    version: int
    precision: Optional[List[str]] = None  # Added in Geti v1.9


@attr.define()
class Score:
    """
    Container class holding a score resulting from a model testing job. The metric
    contained can either relate to a single label (`label_id` will be assigned) or
    averaged over the dataset as a whole (`label_id` will be None)

    Score values range from 0 to 1
    """

    name: str
    value: float
    label_id: Optional[str] = None


@attr.define()
class TestResult:
    """
    Representation of the results of a model test job that was run for a specific
    model and dataset in an Intel® Geti™ project
    """

    datasets_info: List[DatasetInfo]
    id: str
    job_info: JobInfo
    model_info: ModelInfo
    name: str
    scores: List[Score]
    creation_time: Optional[str] = attr.field(default=None, converter=str_to_datetime)

    def get_mean_score(self) -> Score:
        """
        Return the mean score computed over the full dataset

        :return: Mean score on the dataset
        """
        if not self.job_info.is_done:
            raise ValueError(
                "Unable to retrieve mean model score, the model testing job is not "
                "finished yet."
            )
        return [score for score in self.scores if score.label_id is None][0]



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import logging
import os
import tarfile
from typing import Optional

from geti_sdk.demos.constants import DEFAULT_DATA_PATH

from .download_helpers import download_file, set_directory_permissions, validate_hash

DEFAULT_MVTEC_PATH = os.path.join(DEFAULT_DATA_PATH, "mvtec")


def is_ad_dataset(target_folder: str = "data") -> bool:
    """
    Check whether a `target_folder` contains an anomaly detection dataset.

    :param target_folder: Directory to check
    :return: True if the directory contains a valid anomaly detection dataset, False
        otherwise
    """
    content = os.listdir(target_folder)
    expected_directories = ["test", "train"]
    expected_files = ["license.txt", "readme.txt"]
    is_dataset = True
    for dirname in expected_directories + expected_files:
        is_dataset &= dirname in content
    return is_dataset


def get_mvtec_dataset_from_path(dataset_path: str = "data") -> str:
    """
    Download the MVTEC AD 'transistor' dataset if not available.

    Note that the MVTec dataset is released under the following licence:

    License:
        MVTec AD dataset is released under the Creative Commons
        Attribution-NonCommercial-ShareAlike 4.0 International License
        (CC BY-NC-SA 4.0)(https://creativecommons.org/licenses/by-nc-sa/4.0/).

    Reference:
        - Paul Bergmann, Kilian Batzner, Michael Fauser, David Sattlegger, Carsten Steger:
          The MVTec Anomaly Detection Dataset: A Comprehensive Real-World Dataset for
          Unsupervised Anomaly Detection; in: International Journal of Computer Vision
          129(4):1038-1059, 2021, DOI: 10.1007/s11263-020-01400-4.
        - Paul Bergmann, Michael Fauser, David Sattlegger, Carsten Steger: MVTec AD —
          A Comprehensive Real-World Dataset for Unsupervised Anomaly Detection;
          in: IEEE/CVF Conference on Computer Vision and Pattern Recognition (CVPR),
          9584-9592, 2019, DOI: 10.1109/CVPR.2019.00982.

    """
    dataset_name = "transistor"
    os.makedirs(dataset_path, exist_ok=True, mode=0o770)
    transistor_dataset_path = os.path.join(dataset_path, dataset_name)
    if os.path.isdir(transistor_dataset_path) and is_ad_dataset(
        transistor_dataset_path
    ):
        logging.info(
            f"MVTEC '{dataset_name}' dataset found at path {transistor_dataset_path}"
        )
        return transistor_dataset_path

    logging.info(
        f"MVTEC '{dataset_name}' dataset was not found at path {dataset_path}. Making "
        f"an attempt to download the data..."
    )
    print(
        "Note that the MVTec Anomaly Detection dataset is released "
        "under the following license:\n"
        "License:\n"
        "  MVTec AD dataset is released under the Creative Commons \n"
        "  Attribution-NonCommercial-ShareAlike 4.0 International License \n"
        "  (CC BY-NC-SA 4.0)(https://creativecommons.org/licenses/by-nc-sa/4.0/).\n"
        "\n"
        "Reference:\n"
        " - Paul Bergmann, Kilian Batzner, Michael Fauser, David Sattlegger, Carsten Steger:\n"
        "   The MVTec Anomaly Detection Dataset: A Comprehensive Real-World Dataset for\n"
        "   Unsupervised Anomaly Detection; in: International Journal of Computer Vision\n"
        "   129(4):1038-1059, 2021, DOI: 10.1007/s11263-020-01400-4.\n"
        " - Paul Bergmann, Michael Fauser, David Sattlegger, Carsten Steger: MVTec AD —\n"
        "   A Comprehensive Real-World Dataset for Unsupervised Anomaly Detection;\n"
        "   in: IEEE/CVF Conference on Computer Vision and Pattern Recognition (CVPR),\n"
        "   9584-9592, 2019, DOI: 10.1109/CVPR.2019.00982."
    )
    archive_name = f"{dataset_name}.tar.xz"
    url = f"https://www.mydrive.ch/shares/38536/3830184030e49fe74747669442f0f282/download/420938166-1629953277/{archive_name}"
    download_file(url, target_folder=dataset_path, check_valid_archive=False)
    archive_path = os.path.join(dataset_path, archive_name)
    validate_hash(
        file_path=archive_path,
        expected_hash="f9c14ab6c802e69899b529da8b9417b319fa39027a0602ec2beeaf2c3a51e5d527248dfdd3c977a0066fb0ed284c6cdb7236e9cc11d06e927e07072496408be3",
    )

    logging.info(f"Extracting the '{dataset_name}' dataset at path {archive_path}...")
    with tarfile.open(archive_path) as tar_file:
        tar_file.extractall(dataset_path)  # nosec B202

    if not is_ad_dataset(transistor_dataset_path):
        raise ValueError(
            "The dataset was downloaded and extracted successfully, but the directory "
            f"content did not match the expected content. Please ensure that the "
            f"dataset directory at {transistor_dataset_path} contains the expected "
            f"'{dataset_name}' dataset."
        )

    # Fix permissions on extracted files
    set_directory_permissions(transistor_dataset_path)

    logging.info("Cleaning up...")
    os.remove(archive_path)
    return transistor_dataset_path


def get_mvtec_dataset(dataset_path: Optional[str] = None) -> str:
    """
    Check if the MVTEC 'transistor' dataset is present at the specified path. If not,
    this method will attempt to download the dataset to the path specified.

    If no path is passed, this method will check or create the default path: the
    folder 'data' in the top level of the geti-sdk package.

    :param dataset_path: Path to check against.
    :param verbose: True to print detailed output, False to check silently
    :return: Path to the COCO dataset
    """
    if dataset_path is None:
        dataset_path = DEFAULT_MVTEC_PATH
    return get_mvtec_dataset_from_path(dataset_path)



# Copyright (C) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import logging
import os
import shutil
import subprocess  # nosec B404
import tempfile
import time
from typing import List, Optional, Union

import cv2
import imageio_ffmpeg
from tqdm.auto import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from geti_sdk.data_models import Prediction
from geti_sdk.deployment import Deployment
from geti_sdk.utils import show_image_with_annotation_scene


def predict_video_from_deployment(
    video_path: Union[str, os.PathLike],
    deployment: Union[Deployment, str, os.PathLike],
    device: str = "CPU",
    preserve_audio: Optional[bool] = True,
) -> Optional[str]:
    """
    Create a video reconstruction with overlaid model predictions.
    This function runs inference on the local machine for every frame in the video.
    The inference results are overlaid on the frames and the output video path will be returned.

    :param video_path: File path to video
    :param deployment: Path to the folder containing the Deployment data, or Deployment instance
    :param device: Device (CPU or GPU) to load the model to. Defaults to 'CPU'
    :param preserve_audio: True to preserve all audio in the original input video. Defaults to True.
        If ffmpeg could not be found, this option is ignored and no audio would be preserved.
    :return: The file path of the output video if generated successfully. Otherwise None.
    """
    retval: Optional[str] = None

    # prepare deployment for running inference
    if isinstance(deployment, (str, os.PathLike)):
        deployment = Deployment.from_folder(deployment)
    elif not isinstance(deployment, Deployment):
        raise ValueError(f"Unable to read deployment {deployment}")

    logging.info("Load inference models")
    deployment.load_inference_models(device=device)

    # Open the video capture, this prepares the video to be ready for reading
    cap = cv2.VideoCapture(video_path)

    if cap is None or not cap.isOpened():
        raise ValueError(f"Unable to read video from {video_path}")

    # Extract original video properties
    frame_width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
    frame_height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
    fps = cap.get(cv2.CAP_PROP_FPS)
    num_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))

    video_duration = num_frames / fps
    logging.info(
        f"Input video contains {num_frames:.1f} frames, "
        f"for a total duration of {video_duration:.1f} seconds"
    )

    t_start = time.time()

    predictions: List[Prediction] = []
    logging.info("Running video prediction... ")
    with logging_redirect_tqdm(tqdm_class=tqdm), tqdm(
        total=num_frames, desc="Predicting"
    ) as progress_bar:
        while cap.isOpened():
            ret, frame = cap.read()
            if ret is True:
                rgb_frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                prediction = deployment.infer(rgb_frame)
                predictions.append((rgb_frame, prediction))
                progress_bar.update(1)
            else:
                break

    cap.release()

    if len(predictions) == num_frames:
        t_prediction = time.time() - t_start
        logging.info(
            f"Prediction completed successfully in {t_prediction:.1f} seconds. "
        )

        # Determine the output video path
        fname, ext = os.path.splitext(video_path)
        output_video_path = os.path.abspath(fname + "_reconstructed" + ext)

        # Create a video writer to be able to save the reconstructed video
        out_video = cv2.VideoWriter(
            filename=output_video_path,
            fourcc=cv2.VideoWriter_fourcc(*"mp4v"),
            fps=fps,
            frameSize=(frame_width, frame_height),
        )

        count = 0
        logging.info("Running video reconstruction... ")
        with logging_redirect_tqdm(tqdm_class=tqdm), tqdm(
            total=num_frames, desc="Reconstructing"
        ) as progress_bar:
            for rgb_frame, prediction in predictions:
                output_frame = show_image_with_annotation_scene(
                    image=rgb_frame, annotation_scene=prediction, show_results=False
                )
                out_video.write(output_frame)
                count += 1
                progress_bar.update(1)
        out_video.release()

        if preserve_audio is True:
            try:
                # audio = ffmpeg.input(video_path).audio
                # video = ffmpeg.input(output_video_path).video
                # # ffmpeg can't use same input/output video file.
                # temp = tempfile.NamedTemporaryFile(suffix=ext, delete=False)
                # out = ffmpeg.output(video, audio, temp.name)
                # out.run(overwrite_output=True, quiet=True)
                # shutil.move(temp.name, output_video_path)
                FFMPEG = imageio_ffmpeg.get_ffmpeg_exe()
                logging.info("Restoring all audio in the original input video")
                # ffmpeg can't use same input/output video file.
                with tempfile.NamedTemporaryFile(suffix=ext) as temp:
                    cmd = [
                        FFMPEG,
                        "-i",
                        output_video_path,  # 1st input
                        "-i",
                        video_path,  # 2nd input
                        "-map",
                        "0:v",  # video from the 1st input(output_video_path)
                        "-map",
                        "1:a?",  # audio from the 2nd input(video_path). '?' to ignore if no audio exists.
                        "-c",
                        "copy",
                        "-y",
                        temp.name,
                    ]
                    p = subprocess.run(  # nosec B603
                        cmd,
                        stdin=subprocess.PIPE,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                    )
                    if p.returncode == 0:
                        shutil.copy(temp.name, output_video_path)
                    else:
                        logging.warning(
                            "Error occurred while processing audio. No audio would be preserved."
                        )

            except RuntimeError:  # FFMPEG binary could not be found.
                logging.warning(
                    "ffmpeg could not be found on your system. No audio would be preserved."
                )

        retval = output_video_path
        t_reconstruction = time.time() - t_prediction - t_start
        logging.info(
            f"Reconstruction completed successfully in {t_reconstruction:.1f} seconds."
        )
        logging.info(f"Output video saved to `{output_video_path}`")
    else:
        logging.warning("Prediction process failed. ")

    return retval



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

"""
Introduction
------------

The `deployment` package allows creating a deployment of any Intel® Geti™ project.
A project deployment can run inference on an image or video frame locally, i.e.
without any connection to the Intel® Geti™ server.

Deployments can be created for both single task and task chain projects alike, the API
is the same in both cases.

Creating a deployment for a project is done through the
:py:class:`~geti_sdk.sc_rest_client.Geti` class, which provides a
convenience method :py:meth:`~geti_sdk.sc_rest_client.Geti.deploy_project`.

The following code snippet shows:

 #. How to create a deployment for a project
 #. How to use it to run local inference for an image
 #. How to save the deployment to disk

.. code-block:: python

   import cv2

   from geti_sdk import Geti

   geti = Geti(
     host="https://0.0.0.0", username="dummy_user", password="dummy_password"
   )

   # Download the model data and create a `Deployment`
   deployment = geti.deploy_project(project_name="dummy_project")

   # Load the inference models for all tasks in the project, for CPU inference
   deployment.load_inference_models(device='CPU')

   # Run inference
   dummy_image = cv2.imread('dummy_image.png')
   dummy_image = cv2.cvtColor(dummy_image, cv2.COLOR_BGR2RGB)
   prediction = deployment.infer(image=dummy_image)

   # Save the deployment to disk
   deployment.save(path_to_folder="deployment_dummy_project")

A saved Deployment can be loaded from its containing folder using the
:py:meth:`~geti_sdk.deployment.deployment.Deployment.from_folder` method, like so:

.. code-block:: python

   from geti_sdk.deployment import Deployment

   local_deployment = Deployment.from_folder("deployment_dummy_project")

Module contents
---------------

.. automodule:: geti_sdk.deployment.deployed_model
   :members:
   :undoc-members:
   :show-inheritance:

.. automodule:: geti_sdk.deployment.deployment
   :members:
   :undoc-members:
   :show-inheritance:
"""

from .deployed_model import DeployedModel
from .deployment import Deployment

__all__ = ["Deployment", "DeployedModel"]



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import logging
from typing import List, Optional, Sequence

from geti_sdk import Geti
from geti_sdk.annotation_readers import AnnotationReader
from geti_sdk.rest_clients import ProjectClient

from .constants import PROJECT_PREFIX
from .finalizers import force_delete_project
from .project_service import ProjectService


def get_or_create_annotated_project_for_test_class(
    project_service: ProjectService,
    annotation_readers: Sequence[AnnotationReader],
    project_name: str,
    project_type: str = "detection",
    enable_auto_train: bool = False,
    learning_parameter_settings: str = "minimal",
    annotation_requirements_first_training: Optional[int] = None,
):
    """
    This function returns an annotated project with `project_name` of type
    `project_type`.

    :param project_service: ProjectService instance to which the project should be added
    :param annotation_readers: List of AnnotationReader instances from which to get the
        annotations. The number of annotation readers must match the number of
        trainable tasks in the project.
    :param project_name: Name of the project
    :param project_type: Type of the project
    :param enable_auto_train: True to turn auto-training on, False to leave it off
    :param learning_parameter_settings: Settings to use for the learning parameters
        during model training. There are three options:
          'minimal'     - Set hyper parameters such that the training time is minimized
                          (i.e. single epoch, low batch size, etc.)
          'default'     - Use default hyper parameter settings
          'reduced_mem' - Reduce the batch size for memory intensive tasks
    :return: Project instance corresponding to the project on the Intel® Geti™ server
    """
    project_exists = project_service.has_project
    labels = [reader.get_all_label_names() for reader in annotation_readers]

    project = project_service.get_or_create_project(
        project_name=project_name, project_type=project_type, labels=labels
    )
    if not project_exists:
        project_service.set_auto_train(False)
        if learning_parameter_settings == "minimal":
            project_service.set_minimal_training_hypers()
        elif learning_parameter_settings == "reduced_mem":
            project_service.set_reduced_memory_hypers()
        elif learning_parameter_settings != "default":
            logging.info(
                f"Invalid learning parameter settings '{learning_parameter_settings}' "
                f"specified, continuing with default hyper parameters."
            )
        if annotation_requirements_first_training is not None:
            project_service.set_auto_training_annotation_requirement(
                required_images=annotation_requirements_first_training
            )

        project_service.add_annotated_media(
            annotation_readers=annotation_readers, n_images=-1
        )
        project_service.set_auto_train(enable_auto_train)
    return project


def remove_all_test_projects(geti: Geti) -> List[str]:
    """
    Removes all projects created in the SDK tests from the server.

    WARNING: This will remove projects without asking for confirmation. Use with
    caution!

    :param geti: Geti instance pointing to the server from which to remove all
        projects created by the SDK test suite
    """
    project_client = ProjectClient(session=geti.session, workspace_id=geti.workspace_id)
    projects_removed: List[str] = []
    for project in project_client.get_all_projects():
        if project.name.startswith(PROJECT_PREFIX):
            force_delete_project(project.name, project_client)
            projects_removed.append(project.name)
    logging.info(f"{len(projects_removed)} test projects were removed from the server.")
    return projects_removed



# Copyright (C) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from typing import Any, Dict

from geti_sdk.data_models import TestResult
from geti_sdk.utils import deserialize_dictionary


class TestResultRESTConverter:
    """
    Class that handles conversion of Intel® Geti™ REST output for test results to
    objects, and vice-versa
    """

    @staticmethod
    def from_dict(result_dict: Dict[str, Any]) -> TestResult:
        """
        Create a TestResult instance from the input dictionary passed in `result_dict`.

        :param result_dict: Dictionary representing a test result on the Intel® Geti™
            server, as returned by the /tests endpoints
        :return: TestResult instance, holding the result data contained in result_dict
        """
        # Need to convert task type to lower case
        task_type = result_dict["model_info"]["task_type"]
        result_dict["model_info"]["task_type"] = task_type.lower()
        return deserialize_dictionary(result_dict, output_type=TestResult)



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import copy
from typing import Any, Dict, List

import attr

from geti_sdk.data_models import Annotation, AnnotationScene
from geti_sdk.data_models.enums import ShapeType
from geti_sdk.data_models.label import ScoredLabel
from geti_sdk.data_models.shapes import Shape
from geti_sdk.data_models.utils import (
    attr_value_serializer,
    remove_null_fields,
    str_to_shape_type,
)

from .annotation_rest_converter import AnnotationRESTConverter


class NormalizedAnnotationRESTConverter(AnnotationRESTConverter):
    """
    Class containing methods for converting annotations in normalized format to
    and from their REST representation

    It is a legacy class to support the annotation format in a normalized coordinate
    system, which was used in SCv1.1 and below
    """

    @staticmethod
    def _normalized_shape_from_dict(
        input_dict: Dict[str, Any], image_width: int, image_height: int
    ) -> Shape:
        """
        Legacy method to convert shapes represented in normalized coordinates to Shape
        objects. This method is used for reading annotations in SCv1.1 or lower format.

        :param input_dict: Dictionary containing the shape in normalized coordinates
        :param image_width: Width of the image to which the shape applies
        :param image_height: Height of the image to which the shape applies
        :return: Shape object corresponding to the input_dict
        """
        coordinate_keys_x = ["x", "width"]
        coordinate_keys_y = ["y", "height"]
        input_copy = copy.deepcopy(input_dict)
        type_ = str_to_shape_type(input_copy.get("type"))
        if type_ != ShapeType.POLYGON:
            denormalized_coordinates: Dict[str, float] = {}
            for key, value in input_copy.items():
                if key in coordinate_keys_x:
                    new_value = value * image_width
                elif key in coordinate_keys_y:
                    new_value = value * image_height
                else:
                    continue
                denormalized_coordinates.update({key: new_value})
            input_copy.update(denormalized_coordinates)
        else:
            points_dicts = input_copy.pop("points")
            points = [
                dict(x=point["x"] * image_width, y=point["y"] * image_height)
                for point in points_dicts
            ]
            input_copy.update({"points": points})
        return AnnotationRESTConverter._shape_from_dict(input_dict=input_copy)

    @staticmethod
    def normalized_annotation_from_dict(
        input_dict: Dict[str, Any], image_width: int, image_height: int
    ) -> Annotation:
        """
        Legacy method that converts a dictionary representing an annotation (in
        normalized coordinates) to an Annotation object

        :param input_dict:
        :param image_width: Width of the image to which the annotation applies
        :param image_height: Height of the image to which the annotation applies
        :return: Annotation object corresponding to input_dict
        """
        input_copy = copy.deepcopy(input_dict)
        labels: List[ScoredLabel] = []
        for label in input_dict["labels"]:
            labels.append(AnnotationRESTConverter._scored_label_from_dict(label))
        shape = NormalizedAnnotationRESTConverter._normalized_shape_from_dict(
            input_dict["shape"], image_width=image_width, image_height=image_height
        )
        input_copy.update({"labels": labels, "shape": shape})
        return Annotation(**input_copy)

    @staticmethod
    def normalized_annotation_scene_from_dict(
        annotation_scene: Dict[str, Any], image_width: int, image_height: int
    ) -> AnnotationScene:
        """
        Legacy method that creates an AnnotationScene object from a dictionary
        returned by the /annotations REST endpoint in GETi versions 1.1 or below

        :param annotation_scene: dictionary representing an AnnotationScene, which
            contains all annotations for a certain media entity
        :param image_width: Width of the image to which the annotation scene applies
        :param image_height: Height of the image to which the annotation scene applies
        :return: AnnotationScene object
        """
        input_copy = copy.deepcopy(annotation_scene)
        annotations: List[Annotation] = []
        for annotation in annotation_scene["annotations"]:
            annotations.append(
                NormalizedAnnotationRESTConverter.normalized_annotation_from_dict(
                    input_dict=annotation,
                    image_width=image_width,
                    image_height=image_height,
                )
            )
        input_copy.update({"annotations": annotations})
        return AnnotationRESTConverter.from_dict(input_copy)

    @staticmethod
    def to_normalized_dict(
        annotation_scene: AnnotationScene,
        image_width: int,
        image_height: int,
        deidentify: bool = True,
    ) -> Dict[str, Any]:
        """
        Convert an AnnotationScene to a dictionary. By default, removes any ID
        fields in the output dictionary

        :param annotation_scene: AnnotationScene object to convert
        :param image_width:
        :param image_height:
        :param deidentify: True to remove any unique database ID fields in the output,
            False to keep these fields. Defaults to True
        :return: Dictionary holding the serialized AnnotationScene data
        """
        if deidentify:
            annotation_scene.deidentify()
        annotation_scene_dict = attr.asdict(
            annotation_scene, recurse=True, value_serializer=attr_value_serializer
        )
        annotations_serialized: List[Dict[str, Any]] = []
        for annotation in annotation_scene.annotations:
            annotation_dict = attr.asdict(
                annotation, recurse=True, value_serializer=attr_value_serializer
            )
            annotation_dict["shape"] = annotation.shape.to_normalized_coordinates(
                image_width=image_width, image_height=image_height
            )
            annotations_serialized.append(annotation_dict)
        annotation_scene_dict["annotations"] = annotations_serialized
        remove_null_fields(annotation_scene_dict)
        return annotation_scene_dict



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import logging
from typing import Optional

from geti_sdk import Geti
from geti_sdk.annotation_readers import DatumAnnotationReader
from geti_sdk.data_models import Project
from geti_sdk.demos.data_helpers.coco_helpers import get_coco_dataset
from geti_sdk.rest_clients import ProjectClient
from geti_sdk.utils import get_task_types_by_project_type

from .utils import ensure_project_is_trained

DEMO_LABELS = ["dog"]
DEMO_PROJECT_TYPE = "detection"
DEMO_PROJECT_NAME = "COCO dog detection"


def create_segmentation_demo_project(
    geti: Geti,
    n_images: int,
    n_annotations: int = -1,
    auto_train: bool = False,
    dataset_path: Optional[str] = None,
    project_name: str = "Segmentation demo",
) -> Project:
    """
    Create a demo project of type 'segmentation', based off the MS COCO dataset.

    This method creates a project with a single 'Segmentation' task, with segmentation
    mask for labels: 'backpack' and 'suitcase'.

    :param geti: Geti instance, representing the GETi server on which
        the project should be created.
    :param n_images: Number of images that should be uploaded. Pass -1 to upload all
        available images in the dataset for the given labels
    :param n_annotations: Number of images that should be annotated. Pass -1 to
        upload annotations for all images.
    :param auto_train: True to set auto-training to True once the project has been
        created and the images have been annotated, False to leave auto-training
        turned off.
    :param dataset_path: Path to the COCO dataset to use as data source. Defaults to
        the 'data' directory in the top level folder of the geti_sdk package. If
        the dataset is not found in the target folder, this method will attempt to
        download it from the internet.
    :param project_name: Name of the project to create
    :return: Project object, holding detailed information about the project that was
        created on the Intel® Geti™ server.
    """
    coco_path = get_coco_dataset(dataset_path)
    logging.info(" ------- Creating segmentation project --------------- ")

    labels_of_interest = ["backpack", "suitcase"]
    project_type = "segmentation"

    # Create annotation reader
    annotation_reader = DatumAnnotationReader(
        base_data_folder=coco_path, annotation_format="coco"
    )
    annotation_reader.filter_dataset(labels=labels_of_interest, criterion="OR")
    # Create project and upload data
    return geti.create_single_task_project_from_dataset(
        project_name=project_name,
        project_type=project_type,
        path_to_images=coco_path,
        annotation_reader=annotation_reader,
        labels=labels_of_interest,
        number_of_images_to_upload=n_images,
        number_of_images_to_annotate=n_annotations,
        enable_auto_train=auto_train,
    )


def create_detection_demo_project(
    geti: Geti,
    n_images: int,
    n_annotations: int = -1,
    auto_train: bool = False,
    dataset_path: Optional[str] = None,
    project_name: str = "Detection demo",
) -> Project:
    """
    Create a demo project of type 'detection', based off the MS COCO dataset.

    This method creates a project with a single 'Detection' task, with labels and
    annotations for 'cell phone' and 'person' objects

    :param geti: Geti instance, representing the GETi server on which
        the project should be created.
    :param n_images: Number of images that should be uploaded. Pass -1 to upload all
        available images in the dataset for the given labels
    :param n_annotations: Number of images that should be annotated. Pass -1 to
        upload annotations for all images.
    :param auto_train: True to set auto-training to True once the project has been
        created and the images have been annotated, False to leave auto-training
        turned off.
    :param dataset_path: Path to the COCO dataset to use as data source. Defaults to
        the 'data' directory in the top level folder of the geti_sdk package. If
        the dataset is not found in the target folder, this method will attempt to
        download it from the internet.
    :param project_name: Name of the project to create
    :return: Project object, holding detailed information about the project that was
        created on the Intel® Geti™ server.
    """
    coco_path = get_coco_dataset(dataset_path)
    logging.info(" ------- Creating detection project --------------- ")

    labels_of_interest = ["cell phone", "person"]
    project_type = "detection"

    # Create annotation reader
    annotation_reader = DatumAnnotationReader(
        base_data_folder=coco_path, annotation_format="coco"
    )
    annotation_reader.filter_dataset(labels=labels_of_interest, criterion="AND")
    # Create project and upload data
    return geti.create_single_task_project_from_dataset(
        project_name=project_name,
        project_type=project_type,
        path_to_images=coco_path,
        annotation_reader=annotation_reader,
        labels=labels_of_interest,
        number_of_images_to_upload=n_images,
        number_of_images_to_annotate=n_annotations,
        enable_auto_train=auto_train,
    )


def create_classification_demo_project(
    geti: Geti,
    n_images: int,
    n_annotations: int = -1,
    auto_train: bool = False,
    dataset_path: Optional[str] = None,
    project_name: str = "Classification demo",
) -> Project:
    """
    Create a demo project of type 'classification', based off the MS COCO dataset.

    This method creates a project with a single 'Classification' task, with labels to
    classify images with different types of animals. The following animals are
    considered: "horse", "cat", "zebra", "bear"

    :param geti: Geti instance, representing the GETi server on which
        the project should be created.
    :param n_images: Number of images that should be uploaded. Pass -1 to upload all
        available images in the dataset for the given labels
    :param n_annotations: Number of images that should be annotated. Pass -1 to
        upload annotations for all images.
    :param auto_train: True to set auto-training to True once the project has been
        created and the images have been annotated, False to leave auto-training
        turned off.
    :param dataset_path: Path to the COCO dataset to use as data source. Defaults to
        the 'data' directory in the top level folder of the geti_sdk package. If
        the dataset is not found in the target folder, this method will attempt to
        download it from the internet.
    :param project_name: Name of the project to create
    :return: Project object, holding detailed information about the project that was
        created on the Intel® Geti™ server.
    """
    coco_path = get_coco_dataset(dataset_path)
    logging.info(" ------- Creating classification project --------------- ")

    labels_of_interest = ["horse", "cat", "zebra", "bear"]
    project_type = "classification"

    # Create annotation reader
    annotation_reader = DatumAnnotationReader(
        base_data_folder=coco_path, annotation_format="coco"
    )
    annotation_reader.filter_dataset(labels=labels_of_interest, criterion="XOR")
    # Create project and upload data
    return geti.create_single_task_project_from_dataset(
        project_name=project_name,
        project_type=project_type,
        path_to_images=coco_path,
        annotation_reader=annotation_reader,
        labels=labels_of_interest,
        number_of_images_to_upload=n_images,
        number_of_images_to_annotate=n_annotations,
        enable_auto_train=auto_train,
    )


def create_detection_to_segmentation_demo_project(
    geti: Geti,
    n_images: int,
    n_annotations: int = -1,
    auto_train: bool = False,
    dataset_path: Optional[str] = None,
    project_name: str = "Animal detection to segmentation demo",
) -> Project:
    """
    Create a demo project of type 'detection_to_segmentation', based off the MS COCO
    dataset.

    This method creates a project with a 'Detection' task, followed by a 'Segmentation'
    task. The detection task has the label 'animal', and the segmentation task has
    annotations for the following species:
        "dog", "cat", "horse", "elephant", "cow", "sheep", "giraffe", "zebra", "bear"


    :param geti: Geti instance, representing the GETi server on which
        the project should be created.
    :param n_images: Number of images that should be uploaded. Pass -1 to upload all
        available images in the dataset for the given labels
    :param n_annotations: Number of images that should be annotated. Pass -1 to
        upload annotations for all images.
    :param auto_train: True to set auto-training to True once the project has been
        created and the images have been annotated, False to leave auto-training
        turned off.
    :param dataset_path: Path to the COCO dataset to use as data source. Defaults to
        the 'data' directory in the top level folder of the geti_sdk package. If
        the dataset is not found in the target folder, this method will attempt to
        download it from the internet.
    :param project_name: Name of the project to create
    :return: Project object, holding detailed information about the project that was
        created on the Intel® Geti™ server.
    """
    coco_path = get_coco_dataset(dataset_path)
    logging.info(" ------- Creating detection -> segmentation project --------------- ")
    animal_labels = ["dog", "cat", "horse", "cow", "sheep"]
    project_type = "detection_to_segmentation"

    label_source_per_task = []
    for task_type in get_task_types_by_project_type(project_type):
        annotation_reader = DatumAnnotationReader(
            base_data_folder=coco_path, annotation_format="coco", task_type=task_type
        )
        annotation_reader.filter_dataset(labels=animal_labels, criterion="OR")
        label_source_per_task.append(annotation_reader)

    # Group the labels for the first annotation reader, so that the detection task
    # will only see a single label
    label_source_per_task[0].group_labels(
        labels_to_group=animal_labels, group_name="animal"
    )

    # Create project and upload data
    return geti.create_task_chain_project_from_dataset(
        project_name=project_name,
        project_type=project_type,
        path_to_images=coco_path,
        label_source_per_task=label_source_per_task,
        number_of_images_to_upload=n_images,
        number_of_images_to_annotate=n_annotations,
        enable_auto_train=auto_train,
    )


def create_detection_to_classification_demo_project(
    geti: Geti,
    n_images: int,
    n_annotations: int = -1,
    auto_train: bool = False,
    dataset_path: Optional[str] = None,
    project_name: str = "Animal detection to classification demo",
) -> Project:
    """
    Create a demo project of type 'detection_to_classification', based off the MS COCO
    dataset.

    This method creates a project with a 'Detection' task, followed by a
    'Classification' task. The detection task has the label 'animal', and the
    classification task discriminates between 'domestic' and 'wild' animals.

    :param geti: Geti instance, representing the GETi server on which
        the project should be created.
    :param n_images: Number of images that should be uploaded. Pass -1 to upload all
        available images in the dataset for the given labels
    :param n_annotations: Number of images that should be annotated. Pass -1 to
        upload annotations for all images.
    :param auto_train: True to set auto-training to True once the project has been
        created and the images have been annotated, False to leave auto-training
        turned off.
    :param dataset_path: Path to the COCO dataset to use as data source. Defaults to
        the 'data' directory in the top level folder of the geti_sdk package. If
        the dataset is not found in the target folder, this method will attempt to
        download it from the internet.
    :param project_name: Name of the project to create
    :return: Project object, holding detailed information about the project that was
        created on the Intel® Geti™ server.
    """
    coco_path = get_coco_dataset(dataset_path)
    logging.info(
        " ------- Creating detection -> classification project --------------- "
    )
    domestic_labels = ["dog", "cat", "horse", "cow", "sheep"]
    wild_labels = ["elephant", "giraffe", "zebra", "bear"]
    animal_labels = domestic_labels + wild_labels

    project_type = "detection_to_classification"

    label_source_per_task = []
    for task_type in get_task_types_by_project_type(project_type):
        annotation_reader = DatumAnnotationReader(
            base_data_folder=coco_path, annotation_format="coco", task_type=task_type
        )
        annotation_reader.filter_dataset(labels=animal_labels, criterion="OR")
        label_source_per_task.append(annotation_reader)

    # Group the labels for the first annotation reader, so that the detection task
    # will only see a single label
    label_source_per_task[0].group_labels(
        labels_to_group=animal_labels, group_name="animal"
    )
    # Group the labels for the second annotation reader
    label_source_per_task[1].group_labels(
        labels_to_group=domestic_labels, group_name="domestic"
    )
    label_source_per_task[1].group_labels(
        labels_to_group=wild_labels, group_name="wild"
    )
    # Create project and upload data
    return geti.create_task_chain_project_from_dataset(
        project_name=project_name,
        project_type=project_type,
        path_to_images=coco_path,
        label_source_per_task=label_source_per_task,
        number_of_images_to_upload=n_images,
        number_of_images_to_annotate=n_annotations,
        enable_auto_train=auto_train,
    )


def ensure_trained_example_project(
    geti: Geti, project_name: str = DEMO_PROJECT_NAME
) -> Project:
    """
    Ensure that the project specified by `project_name` exists on the GETi
    instance addressed by `geti`.

    :param geti: Geti instance pointing to the GETi server
    :param project_name: Name of the project
    :return: Project object, representing the project in GETi
    """
    project_client = ProjectClient(session=geti.session, workspace_id=geti.workspace_id)
    project = project_client.get_project_by_name(project_name=project_name)

    if project is None:
        # There are two options: Either the user has used the default project name
        # but has not run the `create_coco_project_single_task.py` example yet, or the
        # user has specified a different project which simply doesn't exist.
        #
        # In the first case, we create the project
        #
        # In the second case, we raise an error stating that the project doesn't exist
        # and should be created first
        if project_name == DEMO_PROJECT_NAME:
            print(
                f"\nThe project `{project_name}` does not exist on the server yet, "
                f"creating it now.... \n"
            )
            coco_path = get_coco_dataset()

            # Create annotation reader
            annotation_reader = DatumAnnotationReader(
                base_data_folder=coco_path, annotation_format="coco"
            )
            annotation_reader.filter_dataset(labels=DEMO_LABELS, criterion="OR")
            # Create project and upload data
            project = geti.create_single_task_project_from_dataset(
                project_name=project_name,
                project_type=DEMO_PROJECT_TYPE,
                path_to_images=coco_path,
                annotation_reader=annotation_reader,
                labels=DEMO_LABELS,
                number_of_images_to_upload=50,
                number_of_images_to_annotate=45,
                enable_auto_train=True,
            )
        else:
            raise ValueError(
                f"The project named `{project_name}` does not exist on the server at "
                f"`{geti.session.config.host}`. Please either create it first, or "
                f"specify an existing project."
            )

    ensure_project_is_trained(geti, project)
    return project



# Copyright (C) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

"""
Introduction
------------

The `benchmarking` package contains the
:py:class:`~geti_sdk.benchmarking.benchmarker.Benchmarker` class, which provides
methods for benchmarking models that are trained and deployed with Intel® Geti™.

For example, benchmarking local inference rates for your project can help in selecting
the model architecture to use for your project, or in assessing the performance of the
hardware available for inference.

Module contents
---------------

.. automodule:: geti_sdk.benchmarking.benchmarker
   :members:
   :undoc-members:
   :show-inheritance:
"""

from .benchmarker import Benchmarker

__all__ = ["Benchmarker"]



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import copy
import logging
from pprint import pformat
from typing import Any, ClassVar, Dict, List, Optional, Sequence, Set, Tuple, Union

import attr
import cv2
import numpy as np

from geti_sdk.data_models.annotations import Annotation
from geti_sdk.data_models.enums import AnnotationKind
from geti_sdk.data_models.label import Label, ScoredLabel
from geti_sdk.data_models.media import MediaInformation
from geti_sdk.data_models.media_identifiers import ImageIdentifier, VideoFrameIdentifier
from geti_sdk.data_models.shapes import (
    Ellipse,
    Polygon,
    Rectangle,
    RotatedRectangle,
    Shape,
)
from geti_sdk.data_models.task_annotation_state import TaskAnnotationState
from geti_sdk.data_models.utils import (
    attr_value_serializer,
    deidentify,
    remove_null_fields,
    round_dictionary,
    str_to_annotation_kind,
    str_to_datetime,
)


@attr.define
class AnnotationScene:
    """
    Representation of an annotation scen for a certain media entity in GETi. An
    annotation scene holds all annotations for that specific media entity.

    :var annotations: List of
        :py:class:`~geti_sdk.data_models.annotations.Annotation`s belonging to the
        media entity
    :var id: unique database ID of the AnnotationScene in GETi
    :var kind: :py:class:`~geti_sdk.data_models.enums.annotation_kind.AnnotationKind`
        of the annotation (Annotation or Prediction)
    :var media_identifier: Identifier of the media entity to which this AnnotationScene
        applies
    :var modified: Date and time at which this AnnotationScene was last modified
    :var labels_to_revisit_full_scene: Optional list of database ID's of the labels
        that may not be up to date and therefore need to be revisited for this
        AnnotationScene.
    :var annotation_state_per_task: Optional dictionary holding the annotation state
        for this AnnotationScene for each task in the project pipeline.
    """

    _identifier_fields: ClassVar[List[str]] = ["id", "modified"]
    _GET_only_fields: ClassVar[List[str]] = ["annotation_state_per_task"]

    annotations: List[Annotation]
    kind: str = attr.field(
        converter=str_to_annotation_kind, default=AnnotationKind.ANNOTATION
    )
    media_identifier: Optional[Union[ImageIdentifier, VideoFrameIdentifier]] = None
    id: Optional[str] = None
    modified: Optional[str] = attr.field(converter=str_to_datetime, default=None)
    labels_to_revisit_full_scene: Optional[List[str]] = None
    annotation_state_per_task: Optional[List[TaskAnnotationState]] = None

    @property
    def has_data(self) -> bool:
        """
        Return True if this AnnotationScene has annotation associated with it.
        """
        return len(self.annotations) > 0

    def deidentify(self) -> None:
        """
        Remove all unique database ID's from the annotationscene and the entities it
        contains.
        """
        deidentify(self)
        self.media_identifier = None
        for annotation in self.annotations:
            annotation.deidentify()

    def prepare_for_post(self) -> None:
        """
        Remove all fields that are not valid for making a POST request to the
        /annotations endpoint.

        :return:
        """
        for field_name in self._GET_only_fields:
            setattr(self, field_name, None)

    def append(self, annotation: Annotation) -> None:
        """
        Add an annotation to the annotation scene.

        :param annotation: Annotation to add
        """
        self.annotations.append(annotation)

    def get_by_shape(self, shape: Shape) -> Optional[Annotation]:
        """
        Return the annotation belonging to a specific shape. Returns None if no
        Annotation is found for the shape.

        :param shape: Shape to return the annotation for
        :return:
        """
        return next(
            (
                annotation
                for annotation in self.annotations
                if annotation.shape == shape
            ),
            None,
        )

    def extend(self, annotations: List[Annotation]):
        """
        Extend the list of annotations in the AnnotationScene with additional entries
        in the `annotations` list.

        :param annotations: List of Annotations to add to the AnnotationScene
        :return:
        """
        for annotation in annotations:
            current_shape_annotation = self.get_by_shape(annotation.shape)
            if current_shape_annotation is not None:
                current_shape_annotation.extend_labels(annotation.labels)
            else:
                self.annotations.append(annotation)

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert the AnnotationScene to a dictionary representation.

        :return: Dictionary holding the annotation scene data
        """
        output_dict = attr.asdict(self, value_serializer=attr_value_serializer)
        return output_dict

    @property
    def overview(self) -> str:
        """
        Return a string that gives an overview of the annotation scene.

        :return: overview string of the annotation scene
        """
        deidentified = copy.deepcopy(self)
        deidentified.deidentify()
        dict_output = deidentified.to_dict()
        for annotation_dict in dict_output["annotations"]:
            shape_dict = annotation_dict["shape"]
            annotation_dict["shape"] = round_dictionary(shape_dict)
        remove_null_fields(dict_output)
        return pformat(dict_output)

    @staticmethod
    def _add_shape_to_mask(
        shape: Shape,
        mask: np.ndarray,
        labels: List[ScoredLabel],
        color: Tuple[int, int, int],
        line_thickness: int,
    ) -> np.ndarray:
        """
        Draw an GETi shape entity `shape` on the pixel level mask `mask`. The shape
        will be drawn in the color specified as R,G,B tuple in `color`, using a line
        thickness `line_thickness` (in pixels).

        If the shape represents classification annotations or predictions, this method
        will print the name of the classification labels in the lower-left corner of
        the mask

        :param shape: Shape to add
        :param mask: Mask to draw the shape on
        :param labels: List of labels belonging to the shape
        :param color: RGB tuple representing the color in which the shape should be
            drawn
        :param line_thickness: Line thickness (in pixels) with which the shape should
            be drawn
        :return: Mask with the shape drawn on it
        """
        image_height, image_width = mask.shape[0:-1]
        if isinstance(shape, (Ellipse, Rectangle)):
            x, y = int(shape.x), int(shape.y)
            width, height = int(shape.width), int(shape.height)
            if isinstance(shape, Ellipse):
                cv2.ellipse(
                    mask,
                    center=(x, y),
                    axes=(width, height),
                    angle=0,
                    color=color,
                    startAngle=0,
                    endAngle=360,
                    thickness=line_thickness,
                )
                cv2.ellipse(
                    mask,
                    center=(x, y),
                    axes=(width, height),
                    angle=0,
                    color=(1, 1, 1),
                    startAngle=0,
                    endAngle=360,
                    thickness=1,
                )
            elif isinstance(shape, Rectangle):
                if not shape.is_full_box(
                    image_width=image_width, image_height=image_height
                ):
                    cv2.rectangle(
                        mask,
                        pt1=(x, y),
                        pt2=(x + width, y + height),
                        color=color,
                        thickness=line_thickness,
                    )
                    cv2.rectangle(
                        mask,
                        pt1=(x, y),
                        pt2=(x + width, y + height),
                        color=(1, 1, 1),
                        thickness=1,
                    )
                else:
                    origin = [int(0.01 * image_width), int(0.99 * image_height)]
                    for label in labels:
                        font = cv2.FONT_HERSHEY_SIMPLEX
                        font_scale = 1
                        cv2.putText(
                            mask,
                            label.name,
                            org=origin,
                            fontFace=font,
                            fontScale=font_scale,
                            color=label.color_tuple,
                            thickness=1,
                        )
                        text_width, text_height = cv2.getTextSize(
                            label.name, font, font_scale, line_thickness
                        )[0]
                        origin[0] += text_width + 2
        elif isinstance(shape, RotatedRectangle):
            shape = shape.to_polygon()
        if isinstance(shape, Polygon):
            contour = shape.points_as_contour()
            cv2.drawContours(
                mask,
                contours=[contour],
                color=color,
                thickness=line_thickness,
                contourIdx=-1,
            )
            cv2.drawContours(
                mask, contours=[contour], color=(1, 1, 1), thickness=1, contourIdx=-1
            )
        return mask

    def as_mask(self, media_information: MediaInformation) -> np.ndarray:
        """
        Convert the shapes in the annotation scene to a mask that can be overlayed on
        an image.

        :param media_information: MediaInformation object containing the width and
            height of the image for which the mask should be generated.
        :return: np.ndarray holding the mask representation of the annotation scene
        """
        image_width = media_information.width
        image_height = media_information.height
        mask = np.zeros((image_height, image_width, 3), dtype=np.uint8)

        for annotation in self.annotations:
            label_to_copy_color = annotation.labels[0]
            line_thickness = 3
            shape = annotation.shape
            mask = self._add_shape_to_mask(
                shape=shape,
                mask=mask,
                labels=annotation.labels,
                color=label_to_copy_color.color_tuple,
                line_thickness=line_thickness,
            )
        return mask

    def get_labels(self) -> List[Label]:
        """
        Return a list of all labels present in the annotation scene.

        :return: List of labels
        """
        labels = set()
        for annotation in self.annotations:
            labels.update(annotation.labels)
        return list(labels)

    def get_label_names(self) -> List[str]:
        """
        Return a list with the unique label names in the annotation scene.

        :return: List of label names
        """
        label_names: Set[str] = set()
        for label in self.get_labels():
            label_names.update([label.name])
        return list(label_names)

    def apply_identifier(
        self, media_identifier: Union[ImageIdentifier, VideoFrameIdentifier]
    ) -> "AnnotationScene":
        """
        Apply a `media_identifier` to the current AnnotationScene instance, such
        that the GETi cluster will recognize this AnnotationScene as belonging to the
        media item identified by the media_identifier.

        This method creates and returns a new AnnotationScene instance. The instance
        on which this method is called remains unmodified.

        :param media_identifier: Image or VideoFrame identifier to apply
        :return: new AnnotationScene instance with the identifiers set according to
            `media_identifier`
        """
        new_annotation = copy.deepcopy(self)
        new_annotation.media_identifier = media_identifier
        new_annotation.id = ""
        new_annotation.modified = None
        for annotation in new_annotation.annotations:
            annotation.id = ""
            annotation.modified = ""
        return new_annotation

    def map_labels(
        self, labels: Sequence[Union[Label, ScoredLabel]]
    ) -> "AnnotationScene":
        """
        Attempt to map the labels found in `labels` to those in the AnnotationScene
        instance. Labels are matched by name. This method will return a new
        AnnotationScene object.

        :param labels: Labels to which the existing labels should be mapped
        :return: AnnotationScene with updated labels, corresponding to those found in
            the `project` (if matching labels were found)
        """
        annotations: List[Annotation] = []
        for annotation in self.annotations:
            annotations.append(annotation.map_labels(labels=labels))
        return AnnotationScene(
            annotations=annotations,
            media_identifier=self.media_identifier,
            modified=self.modified,
        )

    def resolve_label_names_and_colors(self, labels: List[Label]) -> None:
        """
        Add label names and colors to all annotations, based on a list of available
        labels.

        :param labels: List of labels for the project, serving as a reference point
            for label names and colors
        """
        name_map = {label.id: label.name for label in labels}
        color_map = {label.id: label.color for label in labels}
        for annotation in self.annotations:
            for label in annotation.labels:
                label.name = name_map.get(label.id, None)
                label.color = color_map.get(label.id, None)
                if label.name is None:
                    logging.warning(
                        f"Unable to resolve label details for label with id "
                        f"`{label.id}`"
                    )



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import os
import re
from importlib import resources
from pathlib import Path
from typing import Tuple

from pathvalidate import sanitize_filepath

from geti_sdk.data_models import OptimizedModel, Project

try:
    OVMS_README_PATH = str(
        resources.files("geti_sdk.deployment.resources") / "OVMS_README.md"
    )
except AttributeError:
    with resources.path("geti_sdk.deployment.resources", "OVMS_README.md") as data_path:
        OVMS_README_PATH = str(data_path)


def generate_ovms_model_name(
    project: Project, model: OptimizedModel, omit_version: bool = True
) -> str:
    """
    Generate a valid name for a model to be uploaded to the OpenVINO Model Server.

    :param project: Project the model belongs to
    :param model: The model for which to generate the name
    :param omit_version: True not to include the version of the model in the name.
        This is useful if the name needs to be used to create a directory structure,
        since including the version will break this.
    :return: String containing the name of the model
    """
    model_name = sanitize_filepath(project.name + "_" + model.name)
    model_name = model_name.replace(" ", "_")
    model_name = model_name.replace("-", "_")
    model_name = model_name.lower()
    if model.version is not None and not omit_version:
        model_name += f":{model.version}"
    return model_name


def generate_ovms_model_address(ovms_address: str, model_name: str) -> str:
    """
    Generate a string representing a valid model address for the OpenVINO Model Server,
    from a given `ovms_address` and `model_name`.

    The format expected for model addresses by OVMS is:

        <address>:<port>/models/<model_name>[:<model_version>]

    Where `address` must not contain the protocol (http or https)

    :param ovms_address: IP address or URL pointing to the OVMS instance, including
        port specification
    :param model_name: Name of the model
    :return: String containing the model address
    """
    model_address = f"{ovms_address}/models/{model_name}"
    if model_address.startswith("https://"):
        model_address = model_address[8:]
    if model_address.startswith("http://"):
        model_address = model_address[7:]
    return model_address


def target_device_is_ovms(device: str) -> bool:
    """
    Return True if the target `device` specified is a URL or IP address, False otherwise

    :param device: Target device string to check
    :return: True if the string represents a URL or IP address, False otherwise
    """
    # Check if 'device' has been specified as a URL or IP address.
    server_pattern = re.compile(
        r"^((https?://)|(www.))(?:([a-zA-Z]+)|(\d+\.\d+\.\d+\.\d+)):\d{1,5}?$"
    )
    return server_pattern.match(device) is not None


def rgb_to_hex(rgb: Tuple[int, int, int]) -> str:
    """
    Convert an RGB color value to its corresponding hexadecimal representation.

    :param rgb: A tuple representing the RGB color value, where each element is an integer between 0 and 255.
    :return: The hexadecimal representation of the RGB color value.

    _Example:

        >>> rgb_to_hex((255, 0, 0))
        '#ff0000'
    """
    return "#{:02x}{:02x}{:02x}".format(rgb[0], rgb[1], rgb[2])


def get_package_version_from_requirements(
    requirements_path: os.PathLike, package_name: str
) -> str:
    """
    Get the version of a package from a requirements file.

    :param requirements_path: The requirements file path
    :param package_name: The name of the package to get the version of
    :return: The version of the package as a string, empty line if the package is not found.
    :raises: ValueError If the requirements file is not found.
    """
    if not package_name:
        return ""

    requirements_path = Path(requirements_path).resolve()
    if not requirements_path.exists():
        raise ValueError(f"Requirements file {requirements_path} not found")

    for line in requirements_path.read_text().split("\n"):
        # The otx may be installed from a GitHub url
        if line.startswith(package_name) and "==" in line:
            return line.split("==")[1].strip()
    return ""



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import logging
import os
import sys
import time
import warnings
from typing import Dict, List, Optional, Sequence, Tuple, Union

import numpy as np
from pathvalidate import sanitize_filepath
from tqdm.auto import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from .annotation_readers import (
    AnnotationReader,
    DatumAnnotationReader,
    GetiAnnotationReader,
)
from .data_models import (
    Dataset,
    Image,
    Prediction,
    Project,
    TaskType,
    Video,
    VideoFrame,
)
from .data_models.containers import MediaList
from .data_models.model import BaseModel
from .deployment import Deployment
from .http_session import (
    GetiRequestException,
    GetiSession,
    ServerCredentialConfig,
    ServerTokenConfig,
)
from .rest_clients import (
    AnnotationClient,
    ConfigurationClient,
    DatasetClient,
    DeploymentClient,
    ImageClient,
    ModelClient,
    PredictionClient,
    ProjectClient,
    VideoClient,
)
from .utils import (
    generate_classification_labels,
    get_default_workspace_id,
    get_project_folder_name,
    get_task_types_by_project_type,
    show_image_with_annotation_scene,
    show_video_frames_with_annotation_scenes,
)

DEFAULT_LOG_LEVEL = logging.INFO
DEFAULT_LOG_FORMAT = "%(asctime)s - %(levelname)s - %(message)s"


class Geti:
    """
    Interact with an Intel® Geti™ server via the REST API.

    The `Geti` class provides methods for project creation, downloading and
    uploading, as well as project deployment. Initializing the class will establish a
    HTTP session to the Intel® Geti™ server, and requires authentication.

    NOTE: The `Geti` instance can either be initialized in the following ways:

        1. Using user credentials. This requires: `host`, `username` and `password` to
            be passed as input
        2. Using a personal access token. This requires: `host` and `token` to be
            passed as input
        3. Using a :py:class:`~geti_sdk.http_session.server_config.ServerTokenConfig`
            or :py:class:`~geti_sdk.http_session.server_config.ServerCredentialConfig`
            instance that contains the full configuration for the Geti server to
            communicate with. This requires `server_config` to be passed as input.

    Arguments for either one of these options must be passed, otherwise a TypeError
    will be raised.

    :param host: IP address or URL at which the cluster can be reached, for example
        'https://0.0.0.0' or 'https://sc_example.intel.com'
    :param username: Username to log in to the cluster
    :param password: Password to log in to the cluster
    :param token: Personal access token that can be used for authentication
    :param workspace_id: Optional ID of the workspace that should be addressed by this
        Geti instance. If not specified, the default workspace is used.
    :param verify_certificate: True to verify the certificate used for making HTTPS
        requests encrypted using TLS protocol. If set to False, an
        InsecureRequestWarning will be issued since in this case requests made to the
        server may be compromised. This should only ever be set to False in a secure
        network environment.
    :param proxies: Optional dictionary containing proxy information. For example
        {
            'http': http://proxy-server.com:<http_port_number>,
            'https': http://proxy-server.com:<https_port_number>
        },
        if set to None (the default), no proxy settings will be used.
    :param server_config: ServerConfiguration instance holding the full details for
        the Geti server to communicate with
    """

    def __init__(
        self,
        host: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        token: Optional[str] = None,
        workspace_id: Optional[str] = None,
        verify_certificate: bool = True,
        proxies: Optional[Dict[str, str]] = None,
        server_config: Optional[
            Union[ServerTokenConfig, ServerCredentialConfig]
        ] = None,
    ):
        # Set up default logging for the SDK.
        if not logging.root.handlers:
            logging.basicConfig(
                handlers=[logging.StreamHandler(stream=sys.stdout)],
                level=DEFAULT_LOG_LEVEL,
                format=DEFAULT_LOG_FORMAT,
            )

        # Validate input parameters
        if host is None and server_config is None:
            raise TypeError(
                "__init__ missing required keyword arguments: Either `host` or "
                "`server_config` must be specified."
            )

        if server_config is None:
            # Set up server configuration with either token or credential authentication
            if token is not None:
                server_config = ServerTokenConfig(
                    host=host,
                    token=token,
                    proxies=proxies,
                    has_valid_certificate=verify_certificate,
                )
                if username is not None or password is not None:
                    warnings.warn(
                        "Both a personal access token and credentials were passed to "
                        "Geti, using token authentication."
                    )
            elif username is not None and password is not None:
                server_config = ServerCredentialConfig(
                    host=host,
                    username=username,
                    password=password,
                    proxies=proxies,
                    has_valid_certificate=verify_certificate,
                )
            else:
                raise TypeError(
                    "__init__ missing required keyword arguments: Either `username` and "
                    "`password` or `token` must be specified."
                )
        else:
            if host is not None:
                warnings.warn(
                    "Both `host` and `server_config` were passed to `Geti`, ignoring "
                    "the value set for `host`."
                )
            if proxies is not None:
                warnings.warn(
                    "Both `proxies` and `server_config` were passed to `Geti`, "
                    "ignoring the value set for `proxies`. If you want to use proxies "
                    "please update the `server_config` accordingly."
                )

        # Initialize session and get workspace id
        self.session = GetiSession(
            server_config=server_config,
        )
        if workspace_id is None:
            workspace_id = get_default_workspace_id(self.session)
        self.workspace_id = workspace_id
        self.project_client = ProjectClient(
            workspace_id=workspace_id, session=self.session
        )

        # Cache of deployment clients for projects in the workspace
        self._deployment_clients: Dict[str, DeploymentClient] = {}

    @property
    def projects(self) -> List[Project]:
        """
        Return a list of projects that are currently available in the workspace on
        the Intel® Geti™ server.

        :return: List of projects in the workspace addressed by the current `Geti`
            instance
        """
        return self.project_client.get_all_projects()

    def get_project(self, project_name: str) -> Project:
        """
        Return the Intel® Geti™ project named `project_name`, if any. If no project by
        that name is found on the Intel® Geti™ server, this method will raise a
        KeyError.

        :param project_name: Name of the project to retrieve
        :raises: KeyError if project named `project_name` is not found on the server
        :return: Project identified by `project_name`
        """
        project = self.project_client.get_project_by_name(project_name=project_name)
        if project is None:
            raise KeyError(
                f"Project '{project_name}' was not found in the current workspace on "
                f"the Intel® Geti™ server."
            )
        return project

    def download_project(
        self,
        project_name: str,
        target_folder: Optional[str] = None,
        include_predictions: bool = False,
        include_active_models: bool = False,
        include_deployment: bool = False,
        max_threads: int = 10,
    ) -> Project:
        """
        Download a project with name `project_name` to the local disk. All images,
        image annotations, videos and video frame annotations in the project are
        downloaded. By default, predictions and models are not downloaded, but they
        can be included by passing `include_predictions=True` and
        `include_active_models=True`, respectively.

        In addition, if `include_deployment` is set to True, this method will create
        and download a deployment for the project as well.

        This method will download data to the path `target_folder`, the contents of the
        folder will be:

            images
                Folder holding all images in the project, if any

            videos
                Folder holding all videos in the project, if any

            annotations
                Directory holding all annotations in the project, in .json format

            predictions
                Directory holding all predictions in the project, in .json format. If
                available, this will include saliency maps in .jpeg format. Only
                created if `include_predictions=True`

            models
                Folder containing the active model for the project. This folder contains
                zip files holding the data for the active models for the tasks in the
                project, and any optimized models derived from them. Models are only
                downloaded if `include_active_models = True`.

            deployment
                Folder containing the deployment for the project, that can be used for
                local inference. The deployment is only created if
                `include_deployment = True`.

            project.json
                File containing the project parameters, that can be used to re-create
                the project.

            configuration.json
                File containing the configurable parameters for the active models in the
                project

        Downloading a project may take a substantial amount of time if the project
        dataset is large.

        :param project_name: Name of the project to download
        :param target_folder: Path to the local folder in which the project data
            should be saved. If not specified, a new directory will be created inside
            the current working directory. The name of the resulting directory will be
            the result of the concatenation of the project unique ID (24 characters)
            and the project name, i.e.: `"{project.id}_{project.name}"`
        :param include_predictions: True to also download the predictions for all
            images and videos in the project, False to not download any predictions.
            If this is set to True but the project has no trained models, downloading
            predictions will be skipped.
        :param include_active_models: True to download the active models for all
            tasks in the project, and any optimized models derived from them. False to
            not download any models. Defaults to False
        :param include_deployment: True to create and download a deployment for the
            project, that can be used for local inference with OpenVINO. Defaults to
            False.
        :param max_threads: Maximum number of threads to use for downloading. Defaults to 10.
            Set to -1 to use all available threads.
        :return: Project object, holding information obtained from the cluster
            regarding the downloaded project
        """
        # Obtain project details from cluster
        project = self.get_project(project_name)

        # Validate or create target_folder
        if target_folder is None:
            target_folder = os.path.join(".", get_project_folder_name(project))
        else:
            sanitize_filepath(target_folder, platform="auto")
        os.makedirs(target_folder, exist_ok=True, mode=0o770)

        # Download project creation parameters:
        self.project_client.download_project_info(
            project_name=project_name, path_to_folder=target_folder
        )

        # Download images
        image_client = ImageClient(
            workspace_id=self.workspace_id, session=self.session, project=project
        )
        images = image_client.get_all_images()
        if len(images) > 0:
            image_client.download_all(
                path_to_folder=target_folder,
                append_image_uid=images.has_duplicate_filenames,
                max_threads=max_threads,
            )

        # Download videos
        video_client = VideoClient(
            workspace_id=self.workspace_id, session=self.session, project=project
        )
        videos = video_client.get_all_videos()
        if len(videos) > 0:
            video_client.download_all(
                path_to_folder=target_folder,
                append_video_uid=videos.has_duplicate_filenames,
                max_threads=max_threads,
            )

        # Download annotations
        annotation_client = AnnotationClient(
            session=self.session, project=project, workspace_id=self.workspace_id
        )
        annotation_client.download_all_annotations(
            path_to_folder=target_folder, max_threads=max_threads
        )

        # Download predictions
        prediction_client = PredictionClient(
            workspace_id=self.workspace_id, session=self.session, project=project
        )
        if prediction_client.ready_to_predict and include_predictions:
            if len(images) > 0:
                prediction_client.download_predictions_for_images(
                    images=images,
                    path_to_folder=target_folder,
                    include_result_media=True,
                )
            if len(videos) > 0:
                prediction_client.download_predictions_for_videos(
                    videos=videos,
                    path_to_folder=target_folder,
                    include_result_media=True,
                    inferred_frames_only=False,
                )

        # Download configuration
        configuration_client = ConfigurationClient(
            workspace_id=self.workspace_id, session=self.session, project=project
        )
        configuration_client.download_configuration(path_to_folder=target_folder)

        # Download active models
        if include_active_models:
            model_client = ModelClient(
                workspace_id=self.workspace_id, session=self.session, project=project
            )
            model_client.download_all_active_models(path_to_folder=target_folder)

        # Download deployment
        if include_deployment:
            logging.info("Creating deployment for project...")
            self.deploy_project(project.name, output_folder=target_folder)

        logging.info(f"Project '{project.name}' was downloaded successfully.")
        return project

    def upload_project(
        self,
        target_folder: str,
        project_name: Optional[str] = None,
        enable_auto_train: bool = True,
        max_threads: int = 5,
    ) -> Project:
        """
        Upload a previously downloaded Intel® Geti™ project to the server. This method
        expects the `target_folder` to contain the following:

            images
                Folder holding all images in the project, if any

            videos
                Folder holding all videos in the project, if any

            annotations
                Directory holding all annotations in the project, in .json format

            project.json
                File containing the project parameters, that can be used to re-create
                the project.

            configuration.json
                Optional file containing the configurable parameters for the active
                models in the project. If this file is not present, the configurable
                parameters for the project will be left at their default values.

        :param target_folder: Folder holding the project data to upload
        :param project_name: Optional name of the project to create on the cluster. If
            left unspecified, the name of the project found in the configuration in
            the `target_folder` will be used.
        :param enable_auto_train: True to enable auto-training for all tasks directly
            after all annotations have been uploaded. This will directly trigger a
            training round if the conditions for auto-training are met. False to leave
            auto-training disabled for all tasks. Defaults to True.
        :param max_threads: Maximum number of threads to use for uploading. Defaults to 5.
            Set to -1 to use all available threads.
        :return: Project object, holding information obtained from the cluster
            regarding the uploaded project
        """
        project = self.project_client.create_project_from_folder(
            path_to_folder=target_folder, project_name=project_name
        )

        # Disable auto-train to prevent the project from training right away
        configuration_client = ConfigurationClient(
            workspace_id=self.workspace_id, session=self.session, project=project
        )
        configuration_client.set_project_auto_train(auto_train=False)

        # Upload media
        image_client = ImageClient(
            workspace_id=self.workspace_id, session=self.session, project=project
        )
        video_client = VideoClient(
            workspace_id=self.workspace_id, session=self.session, project=project
        )

        # Check the media folders inside the project folder. If they are organized
        # according to the projects datasets, upload the media into their corresponding
        # dataset. Otherwise, upload all media into training dataset.
        dataset_client = DatasetClient(
            workspace_id=self.workspace_id, session=self.session, project=project
        )
        if len(project.datasets) == 1 or not dataset_client.has_dataset_subfolders(
            target_folder
        ):
            # Upload all media directly to the training dataset
            images = image_client.upload_folder(
                path_to_folder=os.path.join(target_folder, "images"),
                max_threads=max_threads,
            )
            videos = video_client.upload_folder(
                path_to_folder=os.path.join(target_folder, "videos"),
                max_threads=max_threads,
            )
        else:
            # Make sure that media is uploaded to the correct dataset
            images: MediaList[Image] = MediaList([])
            videos: MediaList[Video] = MediaList([])
            for dataset in project.datasets:
                images.extend(
                    image_client.upload_folder(
                        path_to_folder=os.path.join(
                            target_folder, "images", dataset.name
                        ),
                        dataset=dataset,
                        max_threads=max_threads,
                    )
                )
                videos.extend(
                    video_client.upload_folder(
                        path_to_folder=os.path.join(
                            target_folder, "videos", dataset.name
                        ),
                        dataset=dataset,
                        max_threads=max_threads,
                    )
                )

        # Short sleep to make sure all uploaded media is processed server side
        time.sleep(5)

        # Upload annotations
        annotation_reader = GetiAnnotationReader(
            base_data_folder=os.path.join(target_folder, "annotations"),
            task_type=None,
        )
        annotation_client = AnnotationClient[GetiAnnotationReader](
            session=self.session,
            project=project,
            workspace_id=self.workspace_id,
            annotation_reader=annotation_reader,
        )
        if len(images) > 0:
            annotation_client.upload_annotations_for_images(
                images=images,
            )
        if len(videos) > 0:
            are_videos_processed = False
            start_time = time.time()
            logging.info(
                "Waiting for the Geti server to process all uploaded videos..."
            )
            while (not are_videos_processed) and (time.time() - start_time < 100):
                # Ensure all uploaded videos are processed by the server
                project_videos = video_client.get_all_videos()
                uploaded_ids = {video.id for video in videos}
                project_video_ids = {video.id for video in project_videos}
                are_videos_processed = uploaded_ids.issubset(project_video_ids)
                time.sleep(1)
            annotation_client.upload_annotations_for_videos(
                videos=videos,
            )

        configuration_file = os.path.join(target_folder, "configuration.json")
        if os.path.isfile(configuration_file):
            result = None
            try:
                result = configuration_client.apply_from_file(
                    path_to_folder=target_folder
                )
            except GetiRequestException:
                logging.warning(
                    f"Attempted to set configuration according to the "
                    f"'configuration.json' file in the project directory, but setting "
                    f"the configuration failed. Probably the configuration specified "
                    f"in '{configuration_file}' does "
                    f"not apply to the default model for one of the tasks in the "
                    f"project. Please make sure to reconfigure the models manually."
                )
            if result is None:
                logging.warning(
                    f"Not all configurable parameters could be set according to the "
                    f"configuration in {configuration_file}. Please make sure to "
                    f"verify model configuration manually."
                )
        configuration_client.set_project_auto_train(auto_train=enable_auto_train)
        logging.info(f"Project '{project.name}' was uploaded successfully.")
        return project

    def create_single_task_project_from_dataset(
        self,
        project_name: str,
        project_type: str,
        path_to_images: str,
        annotation_reader: AnnotationReader,
        labels: Optional[List[Union[str, dict]]] = None,
        number_of_images_to_upload: int = -1,
        number_of_images_to_annotate: int = -1,
        enable_auto_train: bool = True,
        upload_videos: bool = False,
        max_threads: int = 5,
    ) -> Project:
        """
        Create a single task project named `project_name` on the Intel® Geti™ server,
        and upload data from a dataset on local disk.

        The type of task that will be in the project can be controlled by setting the
        `project_type`, options are:

            * classification
            * detection
            * segmentation
            * anomaly_classification
            * anomaly_detection
            * anomaly_segmentation
            * instance_segmentation
            * rotated_detection

        If a project called `project_name` exists on the server, this method will
        attempt to upload the media and annotations to the existing project.

        :param project_name: Name of the project to create
        :param project_type: Type of the project, this determines which task the
            project will perform. See above for possible values
        :param path_to_images: Path to the folder holding the images on the local disk.
            See above for details.
        :param annotation_reader: AnnotationReader instance that will be used to
            obtain annotations for the images.
        :param labels: Optional list of labels to use. This will only be used if the
            `annotation_reader` that is passed also supports dataset filtering. If
            not specified, all labels that are found in the dataset are used.
        :param number_of_images_to_upload: Optional integer specifying how many images
            should be uploaded. If not specified, all images found in the dataset are
            uploaded.
        :param number_of_images_to_annotate: Optional integer specifying how many
            images should be annotated. If not specified, annotations for all images
            that have annotations available will be uploaded.
        :param enable_auto_train: True to enable auto-training for all tasks directly
            after all annotations have been uploaded. This will directly trigger a
            training round if the conditions for auto-training are met. False to leave
            auto-training disabled for all tasks. Defaults to True.
        :param upload_videos: True to upload any videos found in the `path_to_images`
            folder.
        :param max_threads: Maximum number of threads to use for uploading. Defaults to 5.
            Set to -1 to use all available threads.
        :return: Project object, holding information obtained from the cluster
            regarding the uploaded project
        """
        if labels is None:
            labels = annotation_reader.get_all_label_names()
        else:
            if project_type == "classification" and not all(
                [isinstance(item, dict) for item in labels]
            ):
                # Handle label generation for classification case
                filter_settings = annotation_reader.applied_filters
                criterion = filter_settings[0]["criterion"]
                multilabel = True
                if criterion == "XOR":
                    multilabel = False
                labels = generate_classification_labels(labels, multilabel=multilabel)
            elif project_type == "anomaly_classification":
                labels = ["Normal", "Anomalous"]

        # Create project
        project = self.project_client.create_project(
            project_name=project_name, project_type=project_type, labels=[labels]
        )
        # Disable auto training
        configuration_client = ConfigurationClient(
            session=self.session, workspace_id=self.workspace_id, project=project
        )
        configuration_client.set_project_auto_train(auto_train=False)

        # Upload images
        image_client = ImageClient(
            session=self.session, workspace_id=self.workspace_id, project=project
        )
        if isinstance(annotation_reader, DatumAnnotationReader):
            if hasattr(annotation_reader, "get_all_image_filepaths"):
                images = image_client.upload_from_list(
                    path_to_folder=path_to_images,
                    image_names=annotation_reader.get_all_image_filepaths(),
                    n_images=number_of_images_to_upload,
                    image_names_as_full_paths=True,
                    max_threads=max_threads,
                )
            else:
                images = image_client.upload_from_list(
                    path_to_folder=path_to_images,
                    image_names=annotation_reader.get_all_image_names(),
                    n_images=number_of_images_to_upload,
                    max_threads=max_threads,
                )
        else:
            images = image_client.upload_folder(
                path_to_images,
                n_images=number_of_images_to_upload,
                max_threads=max_threads,
            )

        if (
            number_of_images_to_annotate < len(images)
            and number_of_images_to_annotate != -1
        ):
            images = images[:number_of_images_to_annotate]

        # Upload videos, if needed
        video_client = VideoClient(
            session=self.session, workspace_id=self.workspace_id, project=project
        )
        videos: MediaList[Video] = MediaList([])
        if upload_videos:
            videos = video_client.upload_folder(
                path_to_folder=path_to_images,
                n_videos=-1,
                max_threads=max_threads,
            )

        # Set annotation reader task type
        annotation_reader.task_type = project.get_trainable_tasks()[0].type
        annotation_reader.prepare_and_set_dataset(
            task_type=project.get_trainable_tasks()[0].type
        )
        # Upload annotations
        annotation_client = AnnotationClient(
            session=self.session,
            project=project,
            workspace_id=self.workspace_id,
            annotation_reader=annotation_reader,
        )
        annotation_client.upload_annotations_for_images(images)

        if len(videos) > 0:
            annotation_client.upload_annotations_for_videos(videos)

        configuration_client.set_project_auto_train(auto_train=enable_auto_train)
        return project

    def create_task_chain_project_from_dataset(
        self,
        project_name: str,
        project_type: str,
        path_to_images: str,
        label_source_per_task: List[Union[AnnotationReader, List[str]]],
        number_of_images_to_upload: int = -1,
        number_of_images_to_annotate: int = -1,
        enable_auto_train: bool = True,
        max_threads: int = 5,
    ) -> Project:
        """
        Create a single task project named `project_name` on the Intel® Geti™ cluster,
        and upload data from a dataset on local disk.

        The type of task that will be in the project can be controlled by setting the
        `project_type`, current options are:

            * detection_to_segmentation
            * detection_to_classification

        If a project called `project_name` exists on the server, this method will
        attempt to upload the media and annotations to the existing project.

        :param project_name: Name of the project to create
        :param project_type: Type of the project, this determines which task the
            project will perform. See above for possible values
        :param path_to_images: Path to the folder holding the images on the local disk.
            See above for details.
        :param label_source_per_task: List containing the label sources for each task
            in the task chain. Each entry in the list corresponds to the label source
            for one task. The list can contain either AnnotationReader instances that
            will be used to obtain the labels for a task, or it can contain a list of
            labels to use for that task.

            For example, in a detection -> classification project we may have labels
            for the first task (for instance 'dog'), but no annotations for the second
            task yet (e.g. ['small', 'large']). In that case the
            `label_source_per_task` should contain:

                [AnnotationReader(), ['small', 'large']]

            Where the annotation reader has been properly instantiated to read the
            annotations for the 'dog' labels.

        :param number_of_images_to_upload: Optional integer specifying how many images
            should be uploaded. If not specified, all images found in the dataset are
            uploaded.
        :param number_of_images_to_annotate: Optional integer specifying how many
            images should be annotated. If not specified, annotations for all images
            that have annotations available will be uploaded.
        :param enable_auto_train: True to enable auto-training for all tasks directly
            after all annotations have been uploaded. This will directly trigger a
            training round if the conditions for auto-training are met. False to leave
            auto-training disabled for all tasks. Defaults to True.
        :param max_threads: Maximum number of threads to use for uploading. Defaults to 5.
            Set to -1 to use all available threads.
        :return: Project object, holding information obtained from the cluster
            regarding the uploaded project
        """
        labels_per_task = [
            (
                entry.get_all_label_names()
                if isinstance(entry, AnnotationReader)
                else entry
            )
            for entry in label_source_per_task
        ]
        annotation_readers_per_task = [
            entry if isinstance(entry, AnnotationReader) else None
            for entry in label_source_per_task
        ]

        task_types = get_task_types_by_project_type(project_type)
        labels_per_task = self._check_unique_label_names(
            labels_per_task=labels_per_task,
            task_types=task_types,
            annotation_readers_per_task=annotation_readers_per_task,
        )

        # Create project
        project = self.project_client.create_project(
            project_name=project_name, project_type=project_type, labels=labels_per_task
        )
        # Disable auto training
        configuration_client = ConfigurationClient(
            session=self.session, workspace_id=self.workspace_id, project=project
        )
        configuration_client.set_project_auto_train(auto_train=False)

        # Upload images
        image_client = ImageClient(
            session=self.session, workspace_id=self.workspace_id, project=project
        )
        # Assume that the first task determines the media that will be uploaded
        first_task_reader = annotation_readers_per_task[0]
        if isinstance(first_task_reader, DatumAnnotationReader):
            images = image_client.upload_from_list(
                path_to_folder=path_to_images,
                image_names=first_task_reader.get_all_image_names(),
                n_images=number_of_images_to_upload,
                max_threads=max_threads,
            )
        else:
            images = image_client.upload_folder(
                path_to_images,
                n_images=number_of_images_to_upload,
                max_threads=max_threads,
            )

        if (
            number_of_images_to_annotate < len(images)
            and number_of_images_to_annotate != -1
        ):
            images = images[:number_of_images_to_annotate]

        append_annotations = False
        previous_task_type = None
        for task_type, reader in zip(task_types, annotation_readers_per_task):
            if reader is not None:
                # Set annotation reader task type
                reader.task_type = task_type
                reader.prepare_and_set_dataset(
                    task_type=task_type, previous_task_type=previous_task_type
                )
                # Upload annotations
                annotation_client = AnnotationClient(
                    session=self.session,
                    project=project,
                    workspace_id=self.workspace_id,
                    annotation_reader=reader,
                )
                annotation_client.upload_annotations_for_images(
                    images=images, append_annotations=append_annotations
                )
                append_annotations = True
            previous_task_type = task_type
        configuration_client.set_project_auto_train(auto_train=enable_auto_train)
        return project

    def download_all_projects(
        self, target_folder: str, include_predictions: bool = True
    ) -> List[Project]:
        """
        Download all projects in the workspace from the Intel® Geti™ server.

        :param target_folder: Directory on local disk to download the project data to.
            If not specified, this method will create a directory named 'projects' in
            the current working directory.
        :param include_predictions: True to also download the predictions for all
            images and videos in the project, False to not download any predictions.
            If this is set to True but the project has no trained models, downloading
            predictions will be skipped.
        :return: List of Project objects, each entry corresponding to one of the
            projects found on the Intel® Geti™ server
        """
        # Obtain project details from cluster
        projects = self.projects

        # Validate or create target_folder
        if target_folder is None:
            target_folder = os.path.join(".", "projects")
        os.makedirs(target_folder, exist_ok=True, mode=0o770)
        logging.info(
            f"Found {len(projects)} projects in the designated workspace on the "
            f"Intel® Geti™ server. Commencing project download..."
        )

        # Download all found projects
        with logging_redirect_tqdm(tqdm_class=tqdm):
            for index, project in enumerate(
                tqdm(projects, desc="Downloading projects")
            ):
                logging.info(
                    f"Downloading project '{project.name}'... {index+1}/{len(projects)}."
                )
                self.download_project(
                    project_name=project.name,
                    target_folder=os.path.join(
                        target_folder, get_project_folder_name(project)
                    ),
                    include_predictions=include_predictions,
                )
        return projects

    def upload_all_projects(self, target_folder: str) -> List[Project]:
        """
        Upload all projects found in the directory `target_folder` on local disk to
        the Intel® Geti™ server.

        This method expects the directory `target_folder` to contain subfolders. Each
        subfolder should correspond to the (previously downloaded) data for one
        project. The method looks for project folders non-recursively, meaning that
        only folders directly below the `target_folder` in the hierarchy are
        considered to be uploaded as project.

        :param target_folder: Directory on local disk to retrieve the project data from
        :return: List of Project objects, each entry corresponding to one of the
            projects uploaded to the Intel® Geti™ server.
        """
        candidate_project_folders = [
            os.path.join(target_folder, subfolder)
            for subfolder in os.listdir(target_folder)
        ]
        project_folders = [
            folder
            for folder in candidate_project_folders
            if ProjectClient.is_project_dir(folder)
        ]
        logging.info(
            f"Found {len(project_folders)} project data folders in the target "
            f"directory '{target_folder}'. Commencing project upload..."
        )
        projects: List[Project] = []
        with logging_redirect_tqdm(tqdm_class=tqdm):
            for index, project_folder in enumerate(
                tqdm(project_folders, desc="Uploading projects")
            ):
                logging.info(
                    f"Uploading project from folder '{os.path.basename(project_folder)}'..."
                    f" {index + 1}/{len(project_folders)}."
                )
                project = self.upload_project(
                    target_folder=project_folder, enable_auto_train=False
                )
                projects.append(project)
        return projects

    def upload_and_predict_media_folder(
        self,
        project_name: str,
        media_folder: str,
        output_folder: Optional[str] = None,
        delete_after_prediction: bool = False,
        skip_if_filename_exists: bool = False,
        max_threads: int = 5,
    ) -> bool:
        """
        Upload a folder with media (images, videos or both) from local disk at path
        `target_folder` to the project with name `project_name` on the Intel® Geti™
        server.
        After the media upload is complete, predictions will be downloaded for all
        media in the folder. This method will create a 'predictions' directory in
        the `target_folder`, containing the prediction output in json format.

        If `delete_after_prediction` is set to True, all uploaded media will be
        removed from the project on the Intel® Geti™ server after the predictions have
        been downloaded.

        :param project_name: Name of the project to upload media to
        :param media_folder: Path to the folder to upload media from
        :param output_folder: Path to save the predictions to. If not specified, this
            method will create a folder named '<media_folder_name>_predictions' on
            the same level as the media_folder
        :param delete_after_prediction: True to remove the media from the project
            once all predictions are received, False to keep the media in the project.
        :param skip_if_filename_exists: Set to True to skip uploading of an image (or
            video) if an image (or video) with the same filename already exists in the
            project. Defaults to False
        :param max_threads: Maximum number of threads to use for uploading. Defaults to 5.
            Set to -1 to use all available threads.
        :return: True if all media was uploaded, and predictions for all media were
            successfully downloaded. False otherwise
        """
        # Obtain project details from cluster
        try:
            project = self.get_project(project_name=project_name)
        except ValueError:
            logging.info(
                f"Project '{project_name}' was not found on the cluster. Aborting "
                f"media upload."
            )
            return False

        # Upload images
        image_client = ImageClient(
            session=self.session, workspace_id=self.workspace_id, project=project
        )
        images = image_client.upload_folder(
            path_to_folder=media_folder,
            skip_if_filename_exists=skip_if_filename_exists,
            max_threads=max_threads,
        )

        # Upload videos
        video_client = VideoClient(
            session=self.session, workspace_id=self.workspace_id, project=project
        )
        videos = video_client.upload_folder(
            path_to_folder=media_folder,
            skip_if_filename_exists=skip_if_filename_exists,
            max_threads=max_threads,
        )

        prediction_client = PredictionClient(
            session=self.session, workspace_id=self.workspace_id, project=project
        )
        if not prediction_client.ready_to_predict:
            logging.info(
                f"Project '{project_name}' is not ready to make predictions, likely "
                f"because one of the tasks in the task chain does not have a "
                f"trained model yet. Aborting prediction."
            )
            return False

        # Set and create output folder if necessary
        if output_folder is None:
            output_folder = media_folder + "_predictions"
        if not os.path.exists(output_folder) and os.path.isdir(output_folder):
            os.makedirs(output_folder, mode=0o770)

        # Request image predictions
        if len(images) > 0:
            prediction_client.download_predictions_for_images(
                images=images, path_to_folder=output_folder
            )

        # Request video predictions
        if len(videos) > 0:
            prediction_client.download_predictions_for_videos(
                videos=videos, path_to_folder=output_folder, inferred_frames_only=False
            )

        # Delete media if required
        result = True
        if delete_after_prediction:
            images_deleted = True
            videos_deleted = True
            if len(images) > 0:
                images_deleted = image_client.delete_images(images=images)
            if len(videos) > 0:
                videos_deleted = video_client.delete_videos(videos=videos)
            result = images_deleted and videos_deleted
        return result

    def upload_and_predict_image(
        self,
        project_name: str,
        image: Union[np.ndarray, Image, VideoFrame, str, os.PathLike],
        visualise_output: bool = True,
        delete_after_prediction: bool = False,
        dataset_name: Optional[str] = None,
    ) -> Tuple[Image, Prediction]:
        """
        Upload a single image to a project named `project_name` on the Intel® Geti™
        server, and return a prediction for it.

        :param project_name: Name of the project to upload the image to
        :param image: Image, numpy array representing an image, or filepath to an
            image to upload and get a prediction for
        :param visualise_output: True to show the resulting prediction, overlayed on
            the image
        :param delete_after_prediction: True to remove the image from the project
            once the prediction is received, False to keep the image in the project.
        :param dataset_name: Optional name of the dataset to which to upload the
            image. The dataset must already exist in the project
        :return: Tuple containing:

            - Image object representing the image that was uploaded
            - Prediction for the image
        """
        project = self.get_project(project_name=project_name)

        # Get the dataset to upload to
        dataset: Optional[Dataset] = None
        if dataset_name is not None:
            dataset_client = DatasetClient(
                session=self.session, workspace_id=self.workspace_id, project=project
            )
            dataset = dataset_client.get_dataset_by_name(dataset_name=dataset_name)

        # Upload the image
        image_client = ImageClient(
            session=self.session, workspace_id=self.workspace_id, project=project
        )
        needs_upload = True
        if isinstance(image, Image):
            if image.id in image_client.get_all_images().ids:
                # Image is already in the project, make sure not to delete it
                needs_upload = False
                image_data = None
            else:
                image_data = image.get_data(self.session)
        else:
            image_data = image
        if needs_upload:
            if image_data is None:
                raise ValueError(
                    f"Cannot upload entity {image}. No data available for upload."
                )
            uploaded_image = image_client.upload_image(
                image=image_data, dataset=dataset
            )
        else:
            uploaded_image = image

        # Get prediction
        prediction_client = PredictionClient(
            session=self.session, workspace_id=self.workspace_id, project=project
        )
        if not prediction_client.ready_to_predict:
            raise ValueError(
                f"Project '{project_name}' is not ready to make predictions. At least "
                f"one of the tasks in the task chain does not have any models trained."
            )
        prediction = prediction_client.get_image_prediction(uploaded_image)
        uploaded_image.get_data(self.session)

        if delete_after_prediction and needs_upload:
            image_client.delete_images(images=MediaList([uploaded_image]))

        if visualise_output:
            show_image_with_annotation_scene(
                image=uploaded_image, annotation_scene=prediction
            )

        return uploaded_image, prediction

    def upload_and_predict_video(
        self,
        project_name: str,
        video: Union[Video, str, os.PathLike, Union[Sequence[np.ndarray], np.ndarray]],
        frame_stride: Optional[int] = None,
        visualise_output: bool = True,
        delete_after_prediction: bool = False,
    ) -> Tuple[Video, MediaList[VideoFrame], List[Prediction]]:
        """
        Upload a single video to a project named `project_name` on the Intel® Geti™
        server, and return a list of predictions for the frames in the video.

        The parameter 'frame_stride' is used to control the stride for frame
        extraction. Predictions are only generated for the extracted frames. So to
        get predictions for all frames, `frame_stride=1` can be passed.

        :param project_name: Name of the project to upload the image to
        :param video: Video or filepath to a video to upload and get predictions for.
            Can also be a 4D numpy array or a list of 3D numpy arrays, shaped such
            that the array dimensions represent `frames x width x height x channels`,
            i.e. each entry holds the pixel data for a video frame.
        :param frame_stride: Frame stride to use. This determines the number of
            frames that will be extracted from the video, and for which predictions
            will be generated
        :param visualise_output: True to show the resulting prediction, overlayed on
            the video frames.
        :param delete_after_prediction: True to remove the video from the project
            once the prediction is received, False to keep the video in the project.
        :return: Tuple containing:

            - Video object holding the data for the uploaded video
            - List of VideoFrames extracted from the video, for which predictions
              have been generated
            - List of Predictions for the Video
        """
        project = self.get_project(project_name=project_name)

        # Upload the video
        video_client = VideoClient(
            session=self.session, workspace_id=self.workspace_id, project=project
        )
        needs_upload = True
        if isinstance(video, Video):
            if video.id in video_client.get_all_videos().ids:
                # Video is already in the project, make sure not to delete it
                needs_upload = False
                video_data = None
            else:
                video_data = video.get_data(self.session)
        elif isinstance(video, (str, os.PathLike)):
            video_data = video
        elif isinstance(video, (Sequence, np.ndarray)):
            if not isinstance(video, np.ndarray):
                video_data = np.array(video)
            else:
                video_data = video
        else:
            video_data = video
        if needs_upload:
            logging.info(f"Uploading video to project '{project_name}'...")
            uploaded_video = video_client.upload_video(video=video_data)
        else:
            uploaded_video = video

        # Get prediction for frames
        prediction_client = PredictionClient(
            session=self.session, workspace_id=self.workspace_id, project=project
        )
        if not prediction_client.ready_to_predict:
            raise ValueError(
                f"Project '{project_name}' is not ready to make predictions. At least "
                f"one of the tasks in the task chain does not have any models trained."
            )
        if frame_stride is None:
            frame_stride = uploaded_video.media_information.frame_stride
        frames = MediaList(
            uploaded_video.to_frames(frame_stride=frame_stride, include_data=True)
        )
        logging.info(
            f"Getting predictions for video '{uploaded_video.name}', using stride "
            f"{frame_stride}"
        )
        predictions = [
            prediction_client.get_video_frame_prediction(frame) for frame in frames
        ]
        if delete_after_prediction and needs_upload:
            video_client.delete_videos(videos=MediaList([uploaded_video]))
        if visualise_output:
            show_video_frames_with_annotation_scenes(
                video_frames=frames, annotation_scenes=predictions
            )
        return uploaded_video, frames, predictions

    def deploy_project(
        self,
        project_name: str,
        output_folder: Optional[Union[str, os.PathLike]] = None,
        models: Optional[Sequence[BaseModel]] = None,
        prepare_ovms_config: bool = False,
    ) -> Deployment:
        """
        Deploy a project by creating a Deployment instance. The Deployment contains
        the optimized models for each task in the project, and can be loaded
        with OpenVINO to run inference locally.

        By default, this method creates a deployment using the current active model
        for each task in the project. However, it is possible to specify a particular
        model to use, by passing it in the list of `models` as input to this method.

        :param project_name: Name of the project to deploy
        :param output_folder: Path to a folder on local disk to which the Deployment
            should be downloaded. If no path is specified, the deployment will not be
            saved.
        :param models: Optional list of models to use in the deployment. This must
            contain at most one model for each task in the project. If for a certain
            task no model is specified, the currently active model for that task will
            be used in the deployment. The order in which the models are passed does
            not matter
        :param prepare_ovms_config: True to prepare the deployment to be hosted on a
            OpenVINO model server (OVMS). Passing True will create OVMS configuration
            files for the model(s) in the project and a README containing the steps to
            launch an OVMS container serving the models.
        :return: Deployment for the project
        """
        project = self.get_project(project_name=project_name)

        deployment_client = self._deployment_clients.get(project.id, None)
        if deployment_client is None:
            # Create deployment client and add to cache.
            deployment_client = DeploymentClient(
                workspace_id=self.workspace_id, session=self.session, project=project
            )
            self._deployment_clients.update({project.id: deployment_client})

        deployment = deployment_client.deploy_project(
            output_folder=output_folder,
            models=models,
            prepare_for_ovms=prepare_ovms_config,
        )
        return deployment

    def logout(self) -> None:
        """
        Log out of the Intel® Geti™ platform and end the HTTP session.
        """
        self.session.logout()

    @staticmethod
    def _check_unique_label_names(
        labels_per_task: List[List[str]],
        task_types: List[TaskType],
        annotation_readers_per_task: List[AnnotationReader],
    ):
        """
        Check that the names of all labels passed in `labels_per_task` are unique. If
        they are not unique and there is a segmentation task in the task chain, this
        method tries to generate segmentation labels in order to guarantee unique label
        names

        :param labels_per_task: Nested list of label names per task
        :param task_types: List of TaskTypes for every trainable task in the project
        :param annotation_readers_per_task: List of annotation readers for all
            trainable tasks in the project
        :raises ValueError: If the label names are not unique and this method is not
            able to generate unique label names for this configuration
        :return: List of labels per task with unique label names
        """
        # Check that label names are unique, try to generate segmentation labels if not
        all_labels = [label for labels in labels_per_task for label in labels]
        if len(set(all_labels)) != len(all_labels):
            new_labels = []
            new_labels_per_task = []
            for index, task_type in enumerate(task_types):
                reader = annotation_readers_per_task[index]
                if task_type == TaskType.SEGMENTATION:
                    if isinstance(reader, DatumAnnotationReader):
                        reader.convert_labels_to_segmentation_names()
                new_labels.extend(reader.get_all_label_names())
                new_labels_per_task.append(reader.get_all_label_names())
            if len(set(new_labels)) != len(new_labels):
                raise ValueError(
                    "Unable to create project. Label names must be unique!"
                )
            else:
                return new_labels_per_task
        else:
            return labels_per_task



# INTEL CONFIDENTIAL
#
# Copyright (C) 2024 Intel Corporation
#
# This software and the related documents are Intel copyrighted materials, and
# your use of them is governed by the express license under which they were provided to
# you ("License"). Unless the License provides otherwise, you may not use, modify, copy,
# publish, distribute, disclose or transmit this software or the related documents
# without Intel's prior written permission.
#
# This software and the related documents are provided as is,
# with no express or implied warranties, other than those that are expressly stated
# in the License.

"""
The module contains utility functions for post-processing predictions.
"""



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import logging
import math
import time
from random import randint
from typing import Dict, List, Sequence, Tuple


def generate_segmentation_labels(detection_labels: List[str]) -> List[str]:
    """
    Generate segmentation label names from a list of detection label names.

    :param detection_labels: label names to generate segmentation label names for
    :return: List of label names
    """
    return [f"{label} shape" for label in detection_labels]


def generate_classification_labels(
    labels: List[str], multilabel: bool = False
) -> List[Dict[str, str]]:
    """
    Generate label creation data from a list of label names. If `multiclass = True`,
    the labels will be generated in such a way that multiple labels can be assigned to
    a single image. If `multiclass = False`, only a single label per image can be
    assigned.

    :param labels: Label names to be used
    :param multilabel: True to be able to assign multiple labels to one image, False
        otherwise. Defaults to False.
    :return: List of dictionaries containing the label data that can be sent to the
        Intel® Geti™ project creation endpoint
    """
    label_list: List[Dict[str, str]] = []
    if multilabel or len(labels) == 1:
        for label in labels:
            label_list.append({"name": label, "group": f"{label}_group"})
    else:
        for label in labels:
            label_list.append({"name": label, "group": "default_classification_group"})
    return label_list


def generate_unique_label_color(label_colors: Sequence[str]) -> str:
    """
    Generate a label color that is unique from the label colors used in `labels`.

    :param label_colors: List of hex color strings with respect to which the new color
        should be generated
    :return: hex string containing the new label color
    """

    def _generate_random_rgb_tuple() -> Tuple[int, int, int]:
        """
        Generate a random R,G,B color tuple. R,G,B are integers on the interval [0,255].

        :return: RGB tuple of integers in [0, 255]
        """
        return randint(0, 255), randint(0, 255), randint(0, 255)  # nosec B311

    def _calculate_rgb_distance(
        color_a: Tuple[int, int, int], color_b: Tuple[int, int, int]
    ) -> float:
        """
        Calculate the root-mean-square difference between two RGB color tuples.
        """
        return math.sqrt(
            (color_a[0] - color_b[0]) ** 2
            + (color_a[1] - color_b[1]) ** 2
            + (color_a[2] - color_b[2]) ** 2
        )

    existing_colors = [
        tuple(int(label[i : i + 2], 16) for i in (1, 3, 5)) for label in label_colors
    ]

    success = False
    t_start = time.time()
    new_color = _generate_random_rgb_tuple()
    distance_threshold = 30 if len(label_colors) < 100 else 10
    while not success and (time.time() - t_start < 100):
        new_color = _generate_random_rgb_tuple()
        min_distance = min(
            [_calculate_rgb_distance(color, new_color) for color in existing_colors]
        )
        if min_distance > distance_threshold:
            success = True
    if not success:
        logging.warning("Unable to generate sufficiently distinct label color.")
    return "#{0:02x}{1:02x}{2:02x}".format(*new_color)



# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

"""The package provides the Visualizer class for models predictions visualization."""


from typing import Optional

import cv2
import numpy as np

from geti_sdk.data_models.annotation_scene import AnnotationScene
from geti_sdk.prediction_visualization.shape_drawer import ShapeDrawer


class Visualizer:
    """
    Visualize the predicted output by drawing the annotations on the input image.

    :example:
        >>> prediction = deployment.infer(image)
        >>> visualizer = Visualizer()
        >>> output_image = visualizer.draw(image, prediction )
        >>> Display(output_image)

    """

    def __init__(
        self,
        window_name: Optional[str] = None,
        show_labels: bool = True,
        show_confidence: bool = True,
        show_count: bool = False,
        is_one_label: bool = False,
        delay: Optional[int] = None,
        output: Optional[str] = None,
    ) -> None:
        """
        Initialize the Visualizer.

        :param window_name: Name of the window to be shown
        :param show_labels: Show labels on the output image
        :param show_confidence: Show confidence on the output image
        :param show_count: Show count of the shapes on the output image
        :param is_one_label: Show only one label on the output image
        :param delay: Delay time for the output image
        :param output: Path to save the output image
        """
        self.window_name = "Window" if window_name is None else window_name
        self.shape_drawer = ShapeDrawer(
            show_count, is_one_label, show_labels, show_confidence
        )

        self.delay = delay
        if delay is None:
            self.delay = 1
        self.output = output

    def draw(
        self,
        image: np.ndarray,
        annotation: AnnotationScene,
        fill_shapes: bool = True,
        confidence_threshold: Optional[float] = None,
        meta: Optional[dict] = None,
    ) -> np.ndarray:
        """
        Draw annotations on the image.

        :param image: Input image in RGB format
        :param annotation: Annotations to be drawn on the input image
        :param meta: Optional meta information
        :param fill_shapes: Fill shapes with color
        :param confidence_threshold: Confidence threshold to filter annotations.
            Must be in range [0, 1].
        :return: Output image with annotations in RGB format
        """
        if confidence_threshold is not None:
            annotation = annotation.filter_by_confidence(confidence_threshold)
        result = self.shape_drawer.draw(
            image, annotation, labels=[], fill_shapes=fill_shapes
        )
        return result

    def show(self, image: np.ndarray) -> None:
        """
        Show result image.

        :param image: Image to be shown (in RGB order).
        """
        image_bgr = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
        cv2.imshow(self.window_name, image_bgr)

    def is_quit(self) -> bool:
        """Check user wish to quit."""
        return ord("q") == cv2.waitKey(self.delay)
